# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import importlib.resources
import json
import logging
import os
import pathlib
import signal
import sys
import threading
import time
import traceback
import warnings
import webbrowser
from typing import Any, Optional, Tuple

import numpy as np
from packaging import version

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Signal
from PySide6.QtGui import QFont, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QFrame,
    QLabel,
    QLineEdit,
    QMainWindow,
    QPushButton,
    QStackedWidget,
    QWidget,
)

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool.a111 import _conf_to_rss_sdk
from acconeer.exptool.a111.algo import Calibration, ModuleInfo
from acconeer.exptool.app import resources

import platformdirs

from . import data_processing
from .elements.helper import (
    CalibrationStatus,
    CalibrationUiState,
    Count,
    ExptoolArgumentParser,
    LoadState,
)
from .elements.modules import (
    MODULE_INFOS,
    MODULE_KEY_TO_MODULE_INFO_MAP,
    MODULE_LABEL_TO_MODULE_INFO_MAP,
)
from .elements.qt_subclasses import (
    AdvancedSerialDialog,
    BiggerMessageBox,
    CollapsibleSection,
    Label,
    SensorSelection,
    SessionInfoView,
)


if "win32" in sys.platform.lower():
    import ctypes

    myappid = "acconeer.exploration.tool"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

HERE = os.path.dirname(os.path.realpath(__file__))
USER_DATA_DIR = platformdirs.user_data_dir(appname="acconeer_exptool", appauthor="Acconeer AB")
LAST_CONF_FILENAME = os.path.join(USER_DATA_DIR, "last_config.npy")

SELECT_A_SERVICE_TEXT = "Select service or detector"

log = logging.getLogger(__name__)


class GUI(QMainWindow):
    sig_scan = Signal(str, str, object)
    sig_sensor_config_pidget_event = Signal(object)
    sig_processing_config_pidget_event = Signal(object)
    sig_calibration_config_pidget_event = Signal(object)

    def __init__(self, under_test=False, use_last_config=True):
        super().__init__()

        self.under_test = under_test
        self.use_last_config = use_last_config

        self.data = None
        self.data_source = None
        self.client = None
        self.num_recv_frames = 0
        self.num_missed_frames = 0
        self.measured_update_rate_fc = et.utils.FreqCounter()
        self.reset_missed_frame_text_time = None
        self.advanced_process_data = {"use_data": False, "process_data": None}
        self.override_baudrate = None
        self.session_info = None
        self.threaded_scan = None

        self.gui_states = {
            "load_state": LoadState.UNLOADED,
            "server_connected": False,
            "replaying_data": False,
            "scan_is_running": False,
            "has_config_error": False,
            "connection_info": None,
        }

        self.current_data_type = None
        self.current_module_label = None
        self.canvas = None
        self.sensors_available = None
        self.calibration: Optional[Calibration] = None
        self.basic_sensor_param_count = Count()
        self.advanced_sensor_param_count = Count()
        self.control_grid_count = Count()
        self.param_grid_count = Count(2)
        self.sensor_widgets = {}

        self.sig_sensor_config_pidget_event.connect(self.pidget_sensor_config_event_handler)
        self.sig_processing_config_pidget_event.connect(
            self.pidget_processing_config_event_handler
        )
        self.sig_calibration_config_pidget_event.connect(
            self.pidget_calibration_config_event_handler
        )

        self.module_label_to_sensor_config_map = {}
        self.module_label_to_processing_config_map = {}
        self.module_label_to_calibration_config_map = {}
        self.current_module_info: Optional[ModuleInfo] = None
        for mi in MODULE_INFOS:
            if mi.sensor_config_class is not None:
                self.module_label_to_sensor_config_map[mi.label] = mi.sensor_config_class()

                processing_config = self.get_default_processing_config(mi.label)
                self.module_label_to_processing_config_map[mi.label] = processing_config

                if mi.calibration_config_class is not None:
                    calibration_config = mi.calibration_config_class()
                    self.module_label_to_calibration_config_map[mi.label] = calibration_config

        with importlib.resources.path(resources, "icon.png") as path:
            self.setWindowIcon(QIcon(str(path)))

        self.main_widget = QtWidgets.QSplitter(self.centralWidget())
        self.main_widget.setStyleSheet("QSplitter::handle{background: lightgrey}")
        self.main_widget.setFrameStyle(QFrame.Panel | QFrame.Sunken)
        self.setCentralWidget(self.main_widget)

        self.init_pyqtgraph()
        self.init_labels()
        self.init_textboxes()
        self.init_buttons()
        self.init_dropdowns()
        self.init_checkboxes()
        self.init_sublayouts()
        self.init_panel_scroll_area()
        self.init_statusbar()

        self.calibration_ui_state = CalibrationUiState(
            load_btn=self.buttons["load_calibration"],
            save_btn=self.buttons["save_calibration"],
            clear_btn=self.buttons["clear_calibration"],
            source_text=self.textboxes["calibration_source"],
            status_label=self.labels["calibration_status"],
            auto_apply_cb=self.checkboxes["calibration_auto_apply"],
            apply_btn=self.buttons["apply_calibration"],
        )
        self.sig_scan.connect(
            lambda start_or_stop, *_: self.calibration_ui_state.set_scan_is_running(
                start_or_stop == "start"
            )
        )

        self.init_pidgets()
        self.update_interface()

        self.canvas_widget = QFrame(self.main_widget)
        self.canvas_layout = QtWidgets.QVBoxLayout(self.canvas_widget)

        self.main_widget.addWidget(self.panel_scroll_area)

        self.update_canvas(force_update=True)

        self.resize(1200, 800)
        self.setWindowTitle("Acconeer Exploration GUI")
        self.show()
        self.start_up()
        self.set_gui_state(None, None)

        self.radar = data_processing.DataProcessing()

        timer = QtCore.QTimer(self)
        timer.timeout.connect(self.plot_timer_fun)
        timer.start(15)
        self.plot_queue = []

    def init_pyqtgraph(self):
        pg.setConfigOption("background", "#f0f0f0")
        pg.setConfigOption("foreground", "k")
        pg.setConfigOption("leftButtonPan", False)
        pg.setConfigOptions(antialias=True)

    def init_labels(self):
        # key: (text)
        label_info = {
            "sensor": ("Sensor",),
            "sweep_buffer": ("Max buffered frames",),
            "data_source": ("",),
            "stored_frames": ("",),
            "interface": ("Interface",),
            "sweep_info": ("",),
            "measured_update_rate": ("",),
            "data_warnings": ("",),
            "rssver": ("",),
            "libver": (f"Lib {et.__version__.split('+')[0]}",),
            "unsupported_mode": ("Mode not supported by this module",),
            "protocol": ("Protocol",),
            "calibration_status": ("",),
        }

        self.labels = {}
        for key, (text,) in label_info.items():
            lbl = QLabel(self)
            lbl.setText(text)
            self.labels[key] = lbl

        for k in ["data_warnings", "unsupported_mode"]:
            lbl = self.labels[k]
            lbl.setStyleSheet("color: #ff0000")
            lbl.setVisible(False)

    def init_textboxes(self):
        # key: (text)
        textbox_info = {
            "host": ("192.168.1.100", True),
            "sweep_buffer": ("1000", True),
            "stored_frames": ("0", False),
            "calibration_source": ("", True),
        }

        self.textboxes = {}
        for key, (text, enabled) in textbox_info.items():
            self.textboxes[key] = QLineEdit(self)
            self.textboxes[key].setText(text)
            self.textboxes[key].setEnabled(enabled)

        calibration_textbox = self.textboxes["calibration_source"]
        calibration_textbox.setReadOnly(True)
        calibration_textbox.setPlaceholderText("No calibration")

    def init_checkboxes(self):
        # text, status, visible, enabled, function, tooltip
        checkbox_info = {
            "verbose": ("Verbose logging", False, True, True, self.set_log_level, None),
            "calibration_auto_apply": (
                "Auto apply calibration",
                False,
                True,
                True,
                self.set_calibration_auto_apply,
                "When checked, any calibration returned from the processor will automatically\n"
                + "be used. It will not apply the buffered calibration.",
            ),
        }

        self.checkboxes = {}
        for key, (text, status, visible, enabled, fun, tooltip) in checkbox_info.items():
            cb = QCheckBox(text, self)
            cb.setChecked(status)
            cb.setVisible(visible)
            cb.setEnabled(enabled)
            cb.setToolTip(tooltip)
            if fun:
                cb.stateChanged.connect(fun)
            self.checkboxes[key] = cb

    def init_graphs(self):
        if self.current_module_info is None:
            with importlib.resources.path(resources, "icon.png") as path:
                canvas = Label(str(path))

            self.refresh_pidgets()
            return canvas

        canvas = pg.GraphicsLayoutWidget()
        self.reload_pg_updater(canvas=canvas)
        self.refresh_pidgets()

        return canvas

    def reload_pg_updater(self, canvas=None, session_info=None):
        if canvas is None:
            canvas = pg.GraphicsLayoutWidget()
            self.swap_canvas(canvas)

        sensor_config = self.get_sensor_config()
        processing_config = self.get_processing_config()

        if session_info is None:
            client = et.a111.Client(mock=True)
            session_info = client.setup_session(sensor_config, check_config=False)

        if self.current_module_info is None:
            return

        self.service_widget = self.current_module_info.pg_updater(
            sensor_config, processing_config, session_info
        )

        self.service_widget.setup(canvas.ci)

    def init_pidgets(self):
        self.last_sensor_config = None

        for sensor_config in self.module_label_to_sensor_config_map.values():
            sensor_config._event_handlers.add(self.pidget_sensor_config_event_handler)
            pidgets = sensor_config._create_pidgets()

            for pidget in pidgets:
                if pidget is None:
                    continue

                category = pidget.param.category
                if category == et.configbase.Category.ADVANCED:
                    grid = self.advanced_sensor_config_section.grid
                    count = self.advanced_sensor_param_count
                else:
                    grid = self.basic_sensor_config_section.grid
                    count = self.basic_sensor_param_count

                grid.addWidget(pidget, count.val, 0, 1, 2)
                count.post_incr()

        self.last_processing_config = None

        for processing_config in self.module_label_to_processing_config_map.values():
            if processing_config is None:
                continue

            processing_config._event_handlers.add(self.pidget_processing_config_event_handler)
            pidgets = processing_config._create_pidgets()

            for pidget in pidgets:
                if pidget is None:
                    continue

                if pidget.param.category == et.configbase.Category.ADVANCED:
                    grid = self.advanced_processing_config_section.grid
                    count = self.param_grid_count
                else:
                    grid = self.basic_processing_config_section.grid
                    count = self.param_grid_count

                grid.addWidget(pidget, count.val, 0, 1, 2)
                count.post_incr()

        self.last_calibration_config = None

        for calibration_config in self.module_label_to_calibration_config_map.values():
            if calibration_config is None:
                continue

            calibration_config._event_handlers.add(self.pidget_calibration_config_event_handler)
            pidgets = calibration_config._create_pidgets()

            for pidget in pidgets:
                if pidget is None:
                    continue

                grid = self.calibration_config_section.grid
                count = self.param_grid_count

                grid.addWidget(pidget, count.val, 0, 1, 2)
                count.post_incr()

        self.refresh_pidgets()
        self.set_gui_state(None, None)

    def refresh_pidgets(self):
        self.refresh_sensor_pidgets()
        self.refresh_processing_pidgets()
        self.refresh_calibration_pidgets()
        self.update_pidgets_on_event()

    def refresh_sensor_pidgets(self):
        sensor_config = self.get_sensor_config()

        if self.last_sensor_config != sensor_config:
            if self.last_sensor_config is not None:
                self.last_sensor_config._state = et.configbase.Config.State.UNLOADED

            self.last_sensor_config = sensor_config

        if sensor_config is None:
            self.basic_sensor_config_section.setVisible(False)
            self.advanced_sensor_config_section.setVisible(False)
            return

        sensor_config._state = et.configbase.Config.State.LOADED

        has_basic_params = has_advanced_params = False
        for param in sensor_config._get_params():
            if param.visible:
                if param.category == et.configbase.Category.ADVANCED:
                    has_advanced_params = True
                else:
                    has_basic_params = True

        self.basic_sensor_config_section.setVisible(has_basic_params)
        self.advanced_sensor_config_section.setVisible(has_advanced_params)

    def refresh_processing_pidgets(self):
        processing_config = self.get_processing_config()

        if self.last_processing_config != processing_config:
            if self.last_processing_config is not None:
                self.last_processing_config._state = et.configbase.Config.State.UNLOADED

            self.last_processing_config = processing_config

        if processing_config is None:
            self.basic_processing_config_section.hide()
            self.advanced_processing_config_section.hide()
            return

        processing_config._state = et.configbase.Config.State.LOADED

        has_basic_params = has_advanced_params = False
        for param in processing_config._get_params():
            if param.visible:
                if param.category == et.configbase.Category.ADVANCED:
                    has_advanced_params = True
                else:
                    has_basic_params = True

        self.basic_processing_config_section.setVisible(has_basic_params)
        self.advanced_processing_config_section.setVisible(has_advanced_params)

    def refresh_calibration_pidgets(self):
        calibration_config = self.get_calibration_config()

        if self.last_calibration_config != calibration_config:
            if self.last_calibration_config is not None:
                self.last_calibration_config._state = et.configbase.Config.State.UNLOADED

            self.last_calibration_config = calibration_config

        if calibration_config is None:
            self.calibration_config_section.hide()
            return

        calibration_config._state = et.configbase.Config.State.LOADED_READONLY

        self.calibration_config_section.setVisible(True)

    def pidget_sensor_config_event_handler(self, sensor_config):
        if threading.current_thread().name != "MainThread":
            self.sig_sensor_config_pidget_event.emit(sensor_config)
            return

        self.update_pidgets_on_event(sensor_config=sensor_config)

    def pidget_processing_config_event_handler(self, processing_config):
        if threading.current_thread().name != "MainThread":
            self.sig_processing_config_pidget_event.emit(processing_config)
            return

        # Processor
        try:
            if isinstance(self.radar.external, self.current_module_info.processor):
                self.radar.external.update_processing_config(processing_config)
        except AttributeError:
            pass

        # Plot updater
        try:
            self.service_widget.update_processing_config(processing_config)
        except AttributeError:
            pass

        self.update_pidgets_on_event(processing_config=processing_config)

    def pidget_calibration_config_event_handler(self, calibration_config):
        if threading.current_thread().name != "MainThread":
            self.sig_calibration_config_pidget_event.emit(calibration_config)
            return

        if self.current_module_info is None or self.current_module_info.calibration_mapper is None:
            return

        # Sync self.calibration
        calib_mapper = self.current_module_info.calibration_mapper
        updated_calibration = calib_mapper.get_updated_calibration_from_configuration(
            configuration=calibration_config,
            calibration=self.calibration,
        )
        self.calibration = updated_calibration

        # Assumes that an update means that the calibration is modified.
        self.calibration_ui_state.modified = True
        self.update_pidgets_on_event(calibration_config=calibration_config)

    def update_pidgets_on_event(
        self, sensor_config=None, processing_config=None, calibration_config=None
    ):
        if sensor_config is None:
            sensor_config = self.get_sensor_config()

        if processing_config is None:
            processing_config = self.get_processing_config()

        if calibration_config is None:
            calibration_config = self.get_calibration_config()

        if sensor_config is None:
            return

        if hasattr(processing_config, "check_sensor_config"):
            pass_on_alerts = processing_config.check_sensor_config(sensor_config)
        else:
            pass_on_alerts = {
                "processing": [],
                "sensor": [],
            }

        all_alerts = []
        alerts = sensor_config._update_pidgets(additional_alerts=pass_on_alerts["sensor"])
        all_alerts.extend(alerts)

        if processing_config is not None:
            alerts = processing_config._update_pidgets(
                additional_alerts=pass_on_alerts["processing"]
            )
            all_alerts.extend(alerts)

        if calibration_config is not None:
            alerts = calibration_config._update_pidgets()
            all_alerts.extend(alerts)

        has_error = any([a.severity == et.configbase.Severity.ERROR for a in all_alerts])
        self.set_gui_state("has_config_error", has_error)

    def init_dropdowns(self):
        self.module_dd = QComboBox(self)
        modules = {}

        for module_info in MODULE_INFOS:
            if module_info.module_family.value not in modules:
                modules[module_info.module_family.value] = [module_info.label]
            else:
                modules[module_info.module_family.value].append(module_info.label)

        self.init_dropdown_sections(modules)

        self.module_dd.currentIndexChanged.connect(self.clear_application_owned_calibration)
        self.module_dd.currentIndexChanged.connect(self.update_canvas)

        self.interface_dd = QComboBox(self)
        self.interface_dd.addItem("Socket", userData=et.a111.Link.SOCKET)
        self.interface_dd.addItem("Serial", userData=et.a111.Link.UART)
        self.interface_dd.addItem("SPI", userData=et.a111.Link.SPI)
        self.interface_dd.addItem("Simulated", userData="mock")
        self.interface_dd.currentIndexChanged.connect(self.update_interface)

        self.ports_dd = QComboBox(self)
        self.ports_dd.hide()
        self.update_ports()

        self.protocol_dd = QComboBox(self)
        self.protocol_dd.setEnabled(False)
        # protocol_dd items and enabled-flag are dynamically set in update_interface.

    def init_dropdown_sections(self, modules: dict):

        self.module_dd.addItem(SELECT_A_SERVICE_TEXT)

        for k, v in modules.items():
            if k is not None:
                if self.module_dd.count() != 0:
                    self.module_dd.insertSeparator(self.module_dd.count())
                    self.module_dd.addItem(k)
                    item = self.module_dd.model().item(self.module_dd.count() - 1, 0)
                    font = QFont()
                    font.setBold(True)
                    item.setFont(font)
                    item.setFlags(item.flags() & ~QtCore.Qt.ItemIsSelectable)

                self.module_dd.addItems(v)

    def set_multi_sensors(self):
        if self.current_module_info is not None:
            module_multi_sensor_support = self.current_module_info.multi_sensor
        else:
            module_multi_sensor_support = False

        if self.get_gui_state("load_state") == LoadState.LOADED:
            source_sensors = json.loads(self.data.sensor_config_dump)["sensor"]
        else:
            source_sensors = self.sensors_available

        for name in self.sensor_widgets:
            self.sensor_widgets[name].set_multi_sensor_support(
                source_sensors, module_multi_sensor_support
            )

    def set_sensors(self, sensors):
        for name in self.sensor_widgets:
            self.sensor_widgets[name].set_sensors(sensors)

    def get_sensors(self, widget_name=None):
        if widget_name is None:
            widget_name = "main"

        sensors = self.sensor_widgets[widget_name].get_sensors()

        return sensors

    def update_ports(self):
        select = -1
        port_tag_tuples = et.utils.get_tagged_serial_ports()
        if os.name == "posix":
            ports = []
            for i, (port, tag) in enumerate(port_tag_tuples):
                tag_string = ""
                if tag:
                    select = i
                    tag_string = " ({})".format(tag)
                ports.append(port + tag_string)
        else:
            ports = [port for port, *_ in port_tag_tuples]

        self.ports_dd.clear()
        self.ports_dd.addItems(ports)
        if select >= 0:
            self.ports_dd.setCurrentIndex(select)

    def advanced_port(self):
        dialog = AdvancedSerialDialog(self.override_baudrate, self)
        ret = dialog.exec()

        if ret == QtWidgets.QDialog.Accepted:
            self.override_baudrate = dialog.get_state()

        dialog.deleteLater()

    def start_btn_clicked(self):
        if self.get_gui_state("load_state") == LoadState.LOADED:
            self.data = None
            self.session_info_view.update(None)
            self.set_gui_state("load_state", LoadState.UNLOADED)
        else:
            self.start_scan()

    def save_file_btn_clicked(self):
        self.save_scan(self.data)

    def replay_btn_clicked(self):
        self.load_scan(restart=True)

    def save_rss_sensor_config_btn_clicked(self):
        stringToSave = _conf_to_rss_sdk.config_to_rss_usage(self, self.get_sensor_config())
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        title = "Save configuration"
        file_types = "Text file (*.txt)"
        filename, info = QtWidgets.QFileDialog.getSaveFileName(
            self, title, "", file_types, options=options
        )

        if filename:
            with open(filename, "w") as f:
                f.write(stringToSave)

    def init_buttons(self):
        # key: text, function, enabled, hidden, group
        button_info = {
            "save_rss_sensor_config": (
                "Save as C SDK config",
                self.save_rss_sensor_config_btn_clicked,
                True,
                False,
                "connection",
            ),
            "start": ("Start", self.start_btn_clicked, False, False, "scan"),
            "connect": ("Connect", self.connect_to_server, True, False, "connection"),
            "stop": ("Stop", self.stop_scan, False, False, "scan"),
            "load_scan": ("Load from file", self.load_scan, True, False, "scan"),
            "save_scan": ("Save to file", self.save_file_btn_clicked, False, False, "scan"),
            "replay_buffered": (
                "Replay",
                self.replay_btn_clicked,
                False,
                False,
                "scan",
            ),
            "scan_ports": ("Scan ports", self.update_ports, True, True, "connection"),
            "sensor_defaults": (
                "Defaults",
                self.sensor_defaults_handler,
                False,
                False,
                "sensor",
            ),
            "service_defaults": (
                "Defaults",
                self.service_defaults_handler,
                True,
                False,
                "service",
            ),
            "service_help": (
                "?",
                self.service_help_button_handler,
                True,
                False,
                "service",
            ),
            "advanced_defaults": (
                "Defaults",
                self.service_defaults_handler,
                True,
                False,
                "advanced",
            ),
            "save_process_data": (
                "Save process data",
                lambda: self.handle_advanced_process_data("save"),
                True,
                True,
                "advanced",
            ),
            "load_process_data": (
                "Load process data",
                lambda: self.handle_advanced_process_data("load"),
                True,
                True,
                "advanced",
            ),
            "advanced_port": (
                "Advanced port settings",
                self.advanced_port,
                True,
                True,
                "connection",
            ),
            "load_calibration": (
                "Load calibration",
                self.calibration_load_button_handler,
                True,
                False,
                "calibration",
            ),
            "save_calibration": (
                "Save calibration",
                self.calibration_save_button_handler,
                True,
                False,
                "calibration",
            ),
            "clear_calibration": (
                "Clear calibration",
                self.calibration_clear_button_handler,
                True,
                False,
                "calibration",
            ),
            "apply_calibration": (
                "Apply calibration",
                self.apply_current_calibration,
                True,
                False,
                "calibration",
            ),
        }

        self.buttons = {}
        for key, (text, fun, enabled, hidden, _) in button_info.items():
            btn = QPushButton(text, self)
            btn.clicked.connect(fun)
            btn.setEnabled(enabled)
            btn.setHidden(hidden)
            btn.setMinimumWidth(150)
            self.buttons[key] = btn

    def init_sublayouts(self):
        self.main_sublayout = QtWidgets.QGridLayout()
        self.main_sublayout.setContentsMargins(0, 3, 0, 3)
        self.main_sublayout.setSpacing(0)

        self.server_section = CollapsibleSection("Connection", is_top=True)
        self.main_sublayout.addWidget(self.server_section, 0, 0)

        # Interface. Always visible
        self.server_section.grid.addWidget(self.labels["interface"], 0, 0)
        self.server_section.grid.addWidget(self.interface_dd, 0, 1)

        # Protocol, always visible.
        self.server_section.grid.addWidget(self.labels["protocol"], 1, 0)
        self.server_section.grid.addWidget(self.protocol_dd, 1, 1)

        # Ports selection. Only visible with UART
        self.server_section.grid.addWidget(self.ports_dd, 2, 0)
        self.server_section.grid.addWidget(self.buttons["scan_ports"], 2, 1)

        # Host textfield. Only visible with Socket
        self.server_section.grid.addWidget(self.textboxes["host"], 3, 0, 1, 2)

        # Advanced port, visible with UART
        self.server_section.grid.addWidget(self.buttons["advanced_port"], 4, 0, 1, 2)

        self.server_section.grid.addWidget(self.buttons["connect"], 5, 0, 1, 2)
        self.server_section.grid.addWidget(self.labels["rssver"], 6, 0, 1, 2)

        self.control_section = CollapsibleSection("Scan controls")
        self.main_sublayout.addWidget(self.control_section, 1, 0)
        c = self.control_grid_count

        # Sublayout for service dropdown and a small help button.
        service_and_help_layout = QtWidgets.QHBoxLayout()
        service_and_help_layout.addWidget(self.module_dd)
        service_and_help_layout.addWidget(self.buttons["service_help"])
        self.buttons["service_help"].setFixedWidth(30)
        self.control_section.grid.addLayout(service_and_help_layout, c.pre_incr(), 0, 1, 2)

        self.control_section.grid.addWidget(self.labels["unsupported_mode"], c.pre_incr(), 0, 1, 2)
        self.control_section.grid.addWidget(self.buttons["start"], c.pre_incr(), 0)
        self.control_section.grid.addWidget(self.buttons["stop"], c.val, 1)
        self.control_section.grid.addWidget(self.buttons["save_scan"], c.pre_incr(), 0)
        self.control_section.grid.addWidget(self.buttons["load_scan"], c.val, 1)
        self.control_section.grid.addWidget(self.buttons["replay_buffered"], c.pre_incr(), 0, 1, 2)
        self.control_section.grid.addWidget(self.labels["data_source"], c.pre_incr(), 0, 1, 2)
        self.control_section.grid.addWidget(self.labels["sweep_buffer"], c.pre_incr(), 0)
        self.control_section.grid.addWidget(self.textboxes["sweep_buffer"], c.val, 1)
        self.control_section.grid.addWidget(self.labels["stored_frames"], c.pre_incr(), 0)
        self.control_section.grid.addWidget(self.textboxes["stored_frames"], c.val, 1)

        self.basic_sensor_config_section = CollapsibleSection("Sensor settings")
        self.main_sublayout.addWidget(self.basic_sensor_config_section, 4, 0)
        c = self.basic_sensor_param_count
        self.basic_sensor_config_section.grid.addWidget(
            self.buttons["sensor_defaults"], c.post_incr(), 0, 1, 2
        )

        self.basic_sensor_config_section.grid.addWidget(
            self.buttons["save_rss_sensor_config"], c.post_incr(), 0, 1, 2
        )

        self.basic_sensor_config_section.grid.addWidget(self.labels["sensor"], c.val, 0)

        sensor_selection = SensorSelection(error_handler=self.error_message)
        self.basic_sensor_config_section.grid.addWidget(sensor_selection, c.post_incr(), 1)
        self.sensor_widgets["main"] = sensor_selection

        self.advanced_sensor_config_section = CollapsibleSection(
            "Advanced sensor settings", init_collapsed=True
        )
        self.main_sublayout.addWidget(self.advanced_sensor_config_section, 5, 0)

        self.session_info_section = CollapsibleSection(
            "Session information (sensor metadata)", init_collapsed=True
        )
        self.main_sublayout.addWidget(self.session_info_section, 6, 0)
        self.session_info_view = SessionInfoView(self.session_info_section)
        self.session_info_section.grid.addWidget(self.session_info_view, 0, 0, 1, 2)

        self.basic_processing_config_section = CollapsibleSection("Processing settings")
        self.main_sublayout.addWidget(self.basic_processing_config_section, 7, 0)
        self.basic_processing_config_section.grid.addWidget(
            self.buttons["service_defaults"], 0, 0, 1, 2
        )

        self.advanced_processing_config_section = CollapsibleSection(
            "Advanced processing settings", init_collapsed=True
        )
        self.main_sublayout.addWidget(self.advanced_processing_config_section, 8, 0)
        self.advanced_processing_config_section.grid.addWidget(
            self.buttons["advanced_defaults"], 0, 0, 1, 2
        )
        self.advanced_processing_config_section.grid.addWidget(
            self.buttons["load_process_data"], 1, 0
        )
        self.advanced_processing_config_section.grid.addWidget(
            self.buttons["save_process_data"], 1, 1
        )

        self.calibration_config_section = CollapsibleSection(
            "Calibration management", init_collapsed=True
        )
        self.main_sublayout.addWidget(self.calibration_config_section, 9, 0)

        self.calibration_config_section.grid.addWidget(QLabel("Calibration source:"), 1, 0)
        self.calibration_config_section.grid.addWidget(self.textboxes["calibration_source"], 1, 1)

        self.calibration_config_section.grid.addWidget(QLabel("Calibration status:"), 2, 0)
        self.calibration_config_section.grid.addWidget(self.labels["calibration_status"], 2, 1)

        self.calibration_config_section.grid.addWidget(self.buttons["load_calibration"], 3, 0)
        self.calibration_config_section.grid.addWidget(self.buttons["save_calibration"], 3, 1)
        self.calibration_config_section.grid.addWidget(
            self.buttons["clear_calibration"], 4, 0, 1, 2
        )
        self.calibration_config_section.grid.addWidget(self.buttons["apply_calibration"], 5, 0)
        self.calibration_config_section.grid.addWidget(
            self.checkboxes["calibration_auto_apply"], 5, 1
        )

        self.main_sublayout.setRowStretch(12, 1)

    def init_panel_scroll_area(self):
        self.panel_scroll_area = QtWidgets.QScrollArea()
        self.panel_scroll_area.setFrameShape(QFrame.NoFrame)
        self.panel_scroll_area.setMinimumWidth(350)
        self.panel_scroll_area.setMaximumWidth(600)
        self.panel_scroll_area.setWidgetResizable(True)
        self.panel_scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.panel_scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.panel_scroll_area.horizontalScrollBar().setEnabled(False)

        self.panel_scroll_area_widget = QStackedWidget(self.panel_scroll_area)
        self.panel_scroll_area.setWidget(self.panel_scroll_area_widget)
        self.main_sublayout_widget = QWidget(self.panel_scroll_area_widget)
        self.main_sublayout_widget.setLayout(self.main_sublayout)
        self.panel_scroll_area_widget.addWidget(self.main_sublayout_widget)
        self.panel_scroll_area_widget.setCurrentWidget(self.main_sublayout_widget)

    def init_statusbar(self):
        self.statusBar().showMessage("Not connected")
        self.labels["sweep_info"].setFixedWidth(220)
        self.labels["sweep_info"].setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.labels["measured_update_rate"].setToolTip("Measured update rate")
        self.labels["measured_update_rate"].setFixedWidth(120)
        self.statusBar().addPermanentWidget(self.labels["data_warnings"])
        self.statusBar().addPermanentWidget(self.labels["sweep_info"])
        self.statusBar().addPermanentWidget(self.labels["measured_update_rate"])
        self.statusBar().addPermanentWidget(self.labels["libver"])
        self.statusBar().addPermanentWidget(self.checkboxes["verbose"])
        self.statusBar().setStyleSheet("QStatusBar{border-top: 1px solid lightgrey;}")
        self.statusBar().show()

    def sensor_defaults_handler(self):
        config = self.get_sensor_config()

        if config is None:
            return

        default_config = self.current_module_info.sensor_config_class()
        config._loads(default_config._dumps())

    def service_defaults_handler(self):
        processing_config = self.get_processing_config()

        if processing_config is not None:
            processing_config._reset()

    def service_help_button_handler(self):
        if self.current_module_info and self.current_module_info.docs_url is None:
            return

        if self.current_module_info is None:
            url = (
                "https://acconeer-python-exploration.readthedocs.io/en/latest/services/index.html"
            )
        else:
            url = self.current_module_info.docs_url

        _ = webbrowser.open_new_tab(url)

    def _current_module_supports_calibration(self) -> bool:
        module_info = self.current_module_info
        return module_info is not None and (
            module_info.calibration_class is not None
            and module_info.calibration_mapper is not None
            and module_info.calibration_config_class is not None
        )

    def calibration_load_button_handler(self):
        """
        Loads a Calibration that was selected in a dialog window.
        The loaded Calibration is then assigned to applied to the current CalibrationConfiguration.
        """
        module_info = self.current_module_info

        if module_info is None:
            self.error_message("Load error", "Please select a module")
            return

        if not self._current_module_supports_calibration():
            self.error_message("Load error", "The selected module does not support calibration")
            return

        file_filter = ";;".join(
            desc for ext, desc in module_info.calibration_class.file_extensions()
        )

        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=self,
            caption="Load calibration",
            filter=file_filter,
            options=QtWidgets.QFileDialog.Options() | QtWidgets.QFileDialog.DontUseNativeDialog,
        )

        if not filename:
            self.error_message("No filename specified.")
            return

        loaded_calibration = module_info.calibration_class.load(path=filename)
        current_calibration_config = self.module_label_to_calibration_config_map[module_info.label]

        module_info.calibration_mapper.update_config_from_calibration(
            configuration=current_calibration_config,
            calibration=loaded_calibration,
        )

        self.calibration = loaded_calibration
        self.calibration_ui_state.load(filename)
        log.info(f"Loaded calibration from {filename}")
        self.refresh_pidgets()

    def calibration_save_button_handler(self):
        module_info = self.current_module_info
        if module_info is None:
            self.error_message("Save Error", "Please select a module")
            return

        if not self._current_module_supports_calibration():
            self.error_message("Save Error", "Selected module does not support calibration")
            return

        if self.calibration is None:
            self.error_message("Save Error", "No calibration is loaded.")
            return

        file_filter = ";;".join(
            desc for ext, desc in module_info.calibration_class.file_extensions()
        )
        exts = [ext for ext, desc in module_info.calibration_class.file_extensions()]

        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption="Save calibration",
            filter=file_filter,
            options=QtWidgets.QFileDialog.Options() | QtWidgets.QFileDialog.DontUseNativeDialog,
        )

        if not filename:
            self.error_message("No filename specified.")
            return

        # if e.g. ".yaml" is not supplied in the save dialog, it won't save as ".yaml" either
        # This adds the first extension that is accepted by the `Calibration`.
        if not filename.endswith("." + exts[0]):
            filename += "." + exts[0]

        self.calibration.save(filename)
        self.calibration_ui_state.save(filename)

        log.info(f"Saved calibration as '{filename}'")

    def clear_application_owned_calibration(self):
        self.calibration_ui_state.clear()
        self.calibration = None

    def calibration_clear_button_handler(self):
        should_clear = not self.calibration_ui_state.modified or self.warning_message(
            "Are you sure you want to clear the calibration?", "Unsaved changes will be lost."
        )
        if not should_clear:
            return

        module_label = self.current_module_label
        calibration_config = self.module_label_to_calibration_config_map.get(module_label)

        if calibration_config:
            calibration_config._reset()

        self.clear_application_owned_calibration()

        processor = self.radar.external
        can_update_processors_calibration = (
            self.current_module_info is not None
            and isinstance(processor, self.current_module_info.processor)
            and hasattr(processor, "update_calibration")
        )
        if can_update_processors_calibration:
            processor.update_calibration(None)
            self.calibration_ui_state.calibration_status = CalibrationStatus.NONE

        self.refresh_pidgets()

    def set_calibration_auto_apply(self):
        cb_is_checked = self.checkboxes["calibration_auto_apply"].isChecked()
        self.calibration_ui_state.auto_apply = cb_is_checked

    def apply_current_calibration(self):
        if not self.calibration:
            self.error_message("Could not apply calibration as it is not present.")
            return

        if isinstance(self.radar.external, self.current_module_info.processor):
            try:
                self.radar.external.update_calibration(self.calibration)
            except Exception as e:
                traceback.print_exc()
                msg = "Failed to apply calibration!\n{}".format(self.format_error(e))
                self.error_message("{}".format(msg))
            else:
                self.calibration_ui_state.calibration_status = CalibrationStatus.IN_PROCESSOR

    def update_canvas(self, force_update=False):
        module_label = self.module_dd.currentText()

        selectable_dd_labels = [SELECT_A_SERVICE_TEXT]
        selectable_dd_labels.extend(MODULE_LABEL_TO_MODULE_INFO_MAP.keys())
        while (
            module_label not in selectable_dd_labels
        ):  # Fixes bug when using arrow keys to go trough dropdown
            direction = self.module_dd.findText(module_label) - self.module_dd.findText(
                self.current_module_label or SELECT_A_SERVICE_TEXT
            )

            if direction < 0:
                index = self.module_dd.findText(module_label) - 2
                # Seperator has an index
            else:
                index = self.module_dd.findText(module_label) + 1

            self.module_dd.blockSignals(True)
            self.module_dd.setCurrentIndex(index)
            self.module_dd.blockSignals(False)

            module_label = self.module_dd.currentText()

        switching_module = self.current_module_label != module_label
        self.current_module_label = module_label

        self.current_module_info = MODULE_LABEL_TO_MODULE_INFO_MAP.get(module_label, None)

        if self.current_module_info is None:
            data_type = None
            self.external = None
        else:
            data_type = self.current_module_info.sensor_config_class().mode
            self.external = self.current_module_info.processor

        switching_data_type = self.current_data_type != data_type
        self.current_data_type = data_type

        if switching_data_type:
            self.data = None
            self.session_info_view.update(None)
            self.set_gui_state("load_state", LoadState.UNLOADED)

        if force_update or switching_module:
            if not switching_module:
                self.get_processing_config()

            new_canvas = self.init_graphs()
            self.swap_canvas(new_canvas)

    def swap_canvas(self, new_canvas):
        if self.canvas is not None:
            self.canvas_layout.removeWidget(self.canvas)
            self.canvas.setParent(None)
            self.canvas.deleteLater()

        self.canvas_layout.addWidget(new_canvas)
        self.canvas = new_canvas

    def populate_protocol_dd(self, *text_data_tuples: Tuple[str, Any]):
        """
        Populates the Protocol dropdown (self.protocol_dd)

        :param text_data_tuples:
            Variable number of tuples of (<displayed text>, <data>).
            The first one passed will be the default (as it will have index=0)
            If only one tuples is passed, the dropdown will be disabled as no choice can be made.
        """
        self.protocol_dd.setEnabled(len(text_data_tuples) > 1)
        for text, data in text_data_tuples:
            self.protocol_dd.addItem(text, userData=data)

    def update_interface(self):
        EXPLORATION_DD_ITEM = ("Exploration", et.a111.Protocol.EXPLORATION)
        AUTODETECT_DD_ITEM = ("Auto-detected", None)
        MODULE_DD_ITEM = ("Module", et.a111.Protocol.MODULE)
        NA_DD_ITEM = ("Not Applicable", None)

        if self.gui_states["server_connected"]:
            self.connect_to_server()
        link = self.interface_dd.currentData()

        # Start from clean slate
        self.ports_dd.hide()
        self.buttons["advanced_port"].hide()
        self.buttons["scan_ports"].hide()
        self.textboxes["host"].hide()
        self.protocol_dd.clear()

        if link == et.a111.Link.UART:
            self.ports_dd.show()
            self.buttons["advanced_port"].show()
            self.buttons["scan_ports"].show()
            self.update_ports()
            self.populate_protocol_dd(MODULE_DD_ITEM, EXPLORATION_DD_ITEM)
        elif link == et.a111.Link.SPI:
            self.populate_protocol_dd(MODULE_DD_ITEM)
        elif link == et.a111.Link.SOCKET:
            self.textboxes["host"].show()
            self.populate_protocol_dd(AUTODETECT_DD_ITEM)
        elif link == "mock":
            self.populate_protocol_dd(NA_DD_ITEM)

        self.set_multi_sensors()

    def format_error(self, e):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        err = "{}\n{}\n{}\n{}".format(exc_type, fname, exc_tb.tb_lineno, e)
        return err

    def error_message(self, text, info_text=None):
        if self.under_test:
            raise Exception

        if not text:
            return

        message_box = BiggerMessageBox(self.main_widget)
        message_box.setIcon(QtWidgets.QMessageBox.Warning)
        message_box.setStandardButtons(QtWidgets.QMessageBox.Ok)
        message_box.setWindowTitle("Alert")

        message_box.setText(text.replace("\n", "<br>"))
        if info_text:
            message_box.setInformativeText(info_text.replace("\n", "<br>"))
        if any(sys.exc_info()):
            detailed_text = traceback.format_exc()
            message_box.setDetailedText(detailed_text)

        message_box.exec()

    def info_handle(self, info, detailed_info=None, blocking=True):
        msg = QtWidgets.QMessageBox(self.main_widget)
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setText(info)
        if detailed_info:
            msg.setDetailedText(detailed_info)
        msg.setWindowTitle("Info")
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        if blocking:
            msg.exec()
        else:
            msg.show()
        return msg

    def warning_message(self, warning, detailed_warning=None):
        msg = QtWidgets.QMessageBox(self.main_widget)
        msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setText(warning)
        if detailed_warning:
            msg.setDetailedText(detailed_warning)
        msg.setWindowTitle("Warning")
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        retval = msg.exec()
        return retval == 1024

    def start_scan(self, from_file=False):
        self.sig_scan.emit("start", "", None)
        if not self.get_sensors():
            self.error_message("Please select at least one sensor")
            return

        if self.current_module_info is None:
            self.error_message("Please select a service or detector")
            return

        try:
            sweep_buffer = int(self.textboxes["sweep_buffer"].text())
            assert sweep_buffer > 0
        except (ValueError, AssertionError):
            self.textboxes["sweep_buffer"].setText(str("1000"))
            self.error_message("Sweep buffer needs to be a positive integer")
            return

        sensor_config = self.get_sensor_config()

        if from_file:
            sensor_config._loads(self.data.sensor_config_dump)
            self.set_gui_state("replaying_data", True)

        self.update_canvas(force_update=True)

        processing_config = self.get_processing_config()
        sensor_config = self.save_gui_settings_to_sensor_config()
        calibration_config = self.get_calibration_config()

        apply_calibration = self.calibration is not None and self.calibration_ui_state.auto_apply

        params = {
            "sensor_config": sensor_config,
            "data_source": "file" if from_file else "stream",
            "module_info": self.current_module_info,
            "sweep_buffer": sweep_buffer,
            "service_params": processing_config,
            "multi_sensor": self.current_module_info.multi_sensor,
            "rss_version": getattr(self, "rss_version", None),
            "calibration": self.calibration if apply_calibration else None,
        }

        self.threaded_scan = Threaded_Scan(params, parent=self)
        self.threaded_scan.sig_scan.connect(self.thread_receive)
        self.sig_scan.connect(self.threaded_scan.receive)

        self.module_dd.setEnabled(False)

        self.num_recv_frames = 0
        self.num_missed_frames = 0
        self.measured_update_rate_fc.reset()
        self.reset_missed_frame_text_time = None
        self.threaded_scan.start()

        if processing_config is not None:
            self.basic_processing_config_section.body_widget.setEnabled(True)
            self.buttons["service_defaults"].setEnabled(False)
            self.buttons["advanced_defaults"].setEnabled(False)
            processing_config._state = et.configbase.Config.State.LIVE
        else:
            self.basic_processing_config_section.body_widget.setEnabled(False)

        if calibration_config:
            calibration_config._state = et.configbase.Config.State.LOADED_READONLY

        if apply_calibration:
            self.calibration_ui_state.calibration_status = CalibrationStatus.IN_PROCESSOR

        self.buttons["connect"].setEnabled(False)

        self.set_gui_state("scan_is_running", True)

    def set_gui_state(self, state, val):
        if state in self.gui_states:
            self.gui_states[state] = val
        elif state is None:
            pass
        else:
            print("{} is an unknown state!".format(state))
            return

        states = self.gui_states

        # Visible, enabled, text

        # Start button
        self.buttons["start"].setEnabled(
            all(
                [
                    self.in_supported_mode or states["load_state"] == LoadState.LOADED,
                    not states["scan_is_running"],
                    not states["has_config_error"],
                    states["server_connected"] or states["load_state"] == LoadState.LOADED,
                ]
            )
        )
        if states["load_state"] == LoadState.LOADED:
            self.buttons["start"].setText("New measurement")
        else:
            self.buttons["start"].setText("Start measurement")

        # Stop button
        self.buttons["stop"].setEnabled(states["scan_is_running"])

        # Save to file button
        self.buttons["save_scan"].setEnabled(
            all(
                [
                    states["load_state"] != LoadState.UNLOADED,
                    not states["scan_is_running"],
                ]
            )
        )

        # Load from file button
        self.buttons["load_scan"].setEnabled(not states["scan_is_running"])

        # Replay button
        self.buttons["replay_buffered"].setEnabled(
            all(
                [
                    states["load_state"] != LoadState.UNLOADED,
                    not states["scan_is_running"],
                ]
            )
        )

        # Data source
        self.labels["data_source"].setVisible(
            bool(states["load_state"] == LoadState.LOADED and self.data_source)
        )
        try:
            text = "Loaded " + os.path.basename(self.data_source)
        except Exception:
            text = ""
        if len(text) > 50:
            text = text[:47] + "..."
        self.labels["data_source"].setText(text)

        # Sweep buffer
        self.labels["sweep_buffer"].setVisible(states["load_state"] != LoadState.LOADED)
        self.textboxes["sweep_buffer"].setVisible(states["load_state"] != LoadState.LOADED)
        self.textboxes["sweep_buffer"].setEnabled(not states["scan_is_running"])

        # Stored frames
        if states["load_state"] == LoadState.LOADED:
            text = "Number of frames"
        else:
            text = "Buffered frames"
        self.labels["stored_frames"].setText(text)
        try:
            num_stored = len(self.data.data)
        except Exception:
            num_stored = 0
        self.textboxes["stored_frames"].setText(str(num_stored))

        # RSS version
        lbl = self.labels["rssver"]

        try:
            strict_ver = states["connection_info"]["strict_version"]
        except Exception:
            strict_ver = None

        if strict_ver is not None:
            if strict_ver < version.parse(et.a111.SDK_VERSION):
                ver_mismatch = "RSS server"
            elif strict_ver > version.parse(et.a111.SDK_VERSION):
                ver_mismatch = "Exploration Tool"
            else:
                ver_mismatch = None

            text = "RSS v" + str(strict_ver)

            if ver_mismatch:
                text += " ({} upgrade recommended)".format(ver_mismatch)
                lbl.setStyleSheet("QLabel {color: red}")
            else:
                lbl.setStyleSheet("")

            lbl.setText(text)
            lbl.show()
        else:
            lbl.hide()

        # Unsupported mode warning
        visible = self.in_supported_mode is not None and not self.in_supported_mode
        self.labels["unsupported_mode"].setVisible(visible)

        # Other

        sensor_config = self.get_sensor_config()
        if sensor_config:
            if states["load_state"] == LoadState.LOADED:
                sensor_config._state = et.configbase.Config.State.LOADED_READONLY
            elif not states["scan_is_running"]:
                sensor_config._state = et.configbase.Config.State.LOADED
            else:
                sensor_config._state = et.configbase.Config.State.LIVE

        calibration_config = self.get_calibration_config()
        if calibration_config:
            if self.calibration is None:
                # Covers case where scan is stopped before a calibration
                # is returned from the processor, which would leave the config
                # in an editable state.
                calibration_config._state = et.configbase.Config.State.LOADED_READONLY
            else:
                if states["load_state"] == LoadState.LOADED:
                    calibration_config._state = et.configbase.Config.State.LOADED_READONLY
                elif not states["scan_is_running"]:
                    calibration_config._state = et.configbase.Config.State.LOADED
                else:
                    calibration_config._state = et.configbase.Config.State.LIVE

        for sensor_widget in self.sensor_widgets.values():
            sensor_widget.setEnabled(not states["scan_is_running"])

        self.set_multi_sensors()
        self.calibration_ui_state.scan_is_running = states["scan_is_running"]

        self.buttons["sensor_defaults"].setEnabled(not states["scan_is_running"])

        if state == "server_connected":
            connected = val
            if connected:
                self.buttons["connect"].setText("Disconnect")
                self.buttons["connect"].setStyleSheet("QPushButton {color: red}")
                self.buttons["advanced_port"].setEnabled(False)
                self.set_multi_sensors()
            else:
                self.buttons["connect"].setText("Connect")
                self.buttons["connect"].setStyleSheet("QPushButton {color: black}")
                self.buttons["advanced_port"].setEnabled(True)
                self.statusBar().showMessage("Not connected")

        if self.current_module_info is None:
            tooltip_text = "Get help with services on the Acconeer documentation pages"
        elif self.current_module_info.docs_url is not None:
            tooltip_text = (
                f'Get help with "{self.current_module_info.label}" '
                + "on the Acconeer documentation pages"
            )
        else:
            tooltip_text = None

        help_button_disabled = (
            self.current_module_info and self.current_module_info.docs_url is None
        )
        self.buttons["service_help"].setEnabled(not help_button_disabled)
        self.buttons["service_help"].setToolTip(tooltip_text)

    def get_gui_state(self, state):
        if state in self.gui_states:
            return self.gui_states[state]
        else:
            print("{} is an unknown state!".format(state))
            return

    def stop_scan(self):
        self.calibration_ui_state.calibration_status = CalibrationStatus.BUFFERED
        self.sig_scan.emit("stop", "", None)

    def unlock_gui(self):
        self.module_dd.setEnabled(True)
        self.buttons["connect"].setEnabled(True)
        self.basic_processing_config_section.body_widget.setEnabled(True)

        processing_config = self.get_processing_config()
        if processing_config is not None:
            self.buttons["service_defaults"].setEnabled(True)
            self.buttons["advanced_defaults"].setEnabled(True)
            processing_config._state = et.configbase.Config.State.LOADED

        self.set_gui_state("replaying_data", False)

        self.set_gui_state("scan_is_running", False)

    def set_log_level(self):
        log_level = logging.INFO
        if self.checkboxes["verbose"].isChecked():
            log_level = logging.DEBUG
        et.utils.set_loglevel(log_level)

    def get_client_arguments_from_ui_elements(self):
        link = self.interface_dd.currentData()
        protocol = self.protocol_dd.currentData()

        if link == "mock":
            return dict(mock=True)

        kwargs = dict(link=link, protocol=protocol)

        if link == et.a111.Link.SOCKET:
            kwargs["host"] = self.textboxes["host"].text()
        elif link == et.a111.Link.UART:
            port_label = self.ports_dd.currentText()
            port, *_ = port_label.split(" ")
            kwargs["serial_port"] = port

        if self.override_baudrate is not None:
            kwargs["override_baudrate"] = self.override_baudrate

        return kwargs

    def connect_to_server(self):
        if not self.get_gui_state("server_connected"):
            client_kwargs = self.get_client_arguments_from_ui_elements()
            self.client = et.a111.Client(**client_kwargs)

            statusbar_connection_info = self.client.description

            if self.override_baudrate:
                log.warning(f"Overriding baudrate ({self.override_baudrate})!")

            self.client.squeeze = False

            try:
                info = self.client.connect()
            except Exception:
                text = "Could not connect to server"
                info_text = None

                if self.client.get_link_type() == et.a111.Link.UART:
                    info_text = (
                        "Did you select the right COM port?"
                        " Try unplugging and plugging back in the module!"
                    )

                self.error_message(text, info_text=info_text)
                return

            self.rss_version = info.get("version_str", None)

            connected_sensors = [1]  # for the initial set
            self.sensors_available = [1]  # for the sensor widget(s)

            if self.client.get_link_type() == et.a111.Link.SOCKET:
                sensor_count = min(info.get("board_sensor_count", 4), 4)
                self.sensors_available = list(range(1, sensor_count + 1))

                if sensor_count > 1:
                    config = et.a111.SparseServiceConfig()
                    connected_sensors = []
                    for i in range(sensor_count):
                        sensor = i + 1
                        config.sensor = sensor
                        try:
                            self.client.start_session(config)
                            self.client.stop_session()
                        except et.a111.SessionSetupError:
                            pass
                        except Exception:
                            self.error_message("Could not connect to server")
                            return
                        else:
                            connected_sensors.append(sensor)
            if self.client.get_link_type() == "mock":
                self.sensors_available = list(range(1, 5))

            if not connected_sensors:
                self.error_message("No sensors connected, check connections")
                try:
                    self.client.disconnect()
                except Exception:
                    pass
                return

            self.set_sensors(connected_sensors)
            self.set_gui_state("server_connected", True)
            self.set_gui_state("load_state", LoadState.UNLOADED)
            self.set_gui_state("connection_info", info)
            self.statusBar().showMessage("Connected via {}".format(statusbar_connection_info))
            if self.current_module_info is None:
                self.module_dd.setCurrentIndex(1)
        else:
            self.sensors_available = None
            self.set_gui_state("server_connected", False)
            self.set_gui_state("connection_info", None)
            self.sig_scan.emit("stop", "", None)
            try:
                self.client.stop_session()
            except Exception:
                pass

            try:
                self.client.disconnect()
            except Exception:
                pass

    def load_gui_settings_from_sensor_config(self, config=None):  # TODO
        return

    def save_gui_settings_to_sensor_config(self):  # TODO
        return self.get_sensor_config()

    def is_float(self, val, is_positive=True):
        try:
            f = float(val)
            if is_positive and f <= 0:
                raise ValueError("Not positive")
            return f
        except Exception:
            return False

    def check_limit(self, val, field, start, end, set_to=None):
        out_of_range = False
        try:
            float(val)
        except (ValueError, TypeError):
            val = start
            out_of_range = True
        if val < start:
            val = start
            out_of_range = True
        if val > end:
            val = end
            out_of_range = True
        if out_of_range:
            if set_to is not None:
                val = set_to
            field.setText(str(val))
        return val, out_of_range

    def load_scan(self, restart=False):
        if restart:
            self.set_gui_state("load_state", LoadState.LOADED)
            self.start_scan(from_file=True)
            return

        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Load scan",
            "",
            "HDF5 data files (*.h5);; NumPy data files (*.npz)",
            options=options,
        )

        if not filename:
            return

        try:
            record = et.a111.recording.load(filename)
        except Exception:
            traceback.print_exc()
            self.error_message(
                (
                    "Failed to load file"
                    "\n\n"
                    "Note: loading data fetched with RSS v1 is not supported."
                    "To load old data, please use an older version of the Exploration Tool."
                )
            )
            return

        try:
            if record.module_key is None:
                raise Exception

            module_info = MODULE_KEY_TO_MODULE_INFO_MAP[record.module_key]
            index = self.module_dd.findText(module_info.label, QtCore.Qt.MatchFixedString)
        except Exception:
            has_loaded_module = False
            print("Can't find the module for the loaded data")
        else:
            has_loaded_module = index != -1

        if not has_loaded_module:  # Just try loading the data in the service-only module
            try:
                module_info = MODULE_KEY_TO_MODULE_INFO_MAP[record.mode.name.lower()]
                index = self.module_dd.findText(module_info.label, QtCore.Qt.MatchFixedString)
            except Exception:
                self.error_message("Unknown mode in loaded file")
                return

        self.module_dd.setCurrentIndex(index)
        self.update_canvas()

        self.data = record

        sensor_config = self.get_sensor_config()
        sensor_config._loads(record.sensor_config_dump)

        # Order is important for the following 3 calls
        self.set_multi_sensors()
        self.set_sensors(sensor_config.sensor)
        self.set_gui_state("load_state", LoadState.LOADED)

        if has_loaded_module:
            processing_config = self.get_processing_config()
            if isinstance(processing_config, et.configbase.ProcessingConfig):
                if record.processing_config_dump is not None:
                    try:
                        processing_config._loads(record.processing_config_dump)
                    except Exception:
                        traceback.print_exc()
            else:
                try:
                    self.load_legacy_processing_config_dump(record)
                except Exception:
                    traceback.print_exc()

        self.data_source = filename
        self.start_scan(from_file=True)

    def save_scan(self, record):
        if len(record.data) == 0:
            self.error_message("No data to save")
            return

        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        title = "Save scan"
        file_types = "HDF5 data files (*.h5);; NumPy data files (*.npz)"
        filename, info = QtWidgets.QFileDialog.getSaveFileName(
            self, title, "", file_types, options=options
        )

        if not filename:
            return

        record.mode = self.get_sensor_config().mode
        record.module_key = self.current_module_info.key

        record.processing_config_dump = None
        record.legacy_processing_config_dump = None

        processing_config = self.get_processing_config()
        if processing_config is not None:
            record.processing_config_dump = processing_config._dumps()

        try:
            if "h5" in info:
                et.a111.recording.save_h5(filename, record)
            else:
                et.a111.recording.save_npz(filename, record)
        except Exception as e:
            traceback.print_exc()
            self.error_message("Failed to save file:\n {:s}".format(e))

    def thread_receive(self, message_type, message, data=None):
        if "error" in message_type:
            if message_type == "session_setup_error":
                error = "Failed to setup session (bad config)!\n"
                if "socket" in self.interface_dd.currentText().lower():
                    error += "Check that selected sensors are connected and working!\n"
                    error += "Check Streaming server log for erros!"
                self.error_message(error)
            elif "client" in message_type:
                self.stop_scan()
                if self.get_gui_state("server_connected"):
                    self.connect_to_server()
            elif "proccessing" in message_type:
                self.stop_scan()
            self.error_message("{}".format(message))
        elif message_type == "scan_data":
            if self.get_gui_state("load_state") != LoadState.LOADED:
                self.data = data
                self.data_source = None
                self.set_gui_state("load_state", LoadState.BUFFERED)
        elif message_type == "scan_done":
            self.unlock_gui()
        elif "update_external_plots" in message_type:
            if data is not None:
                self.update_external_plots(data)
        elif "sweep_info" in message_type:
            self.update_sweep_info(data)
        elif "session_info" in message_type:
            self.session_info = data
            self.reload_pg_updater(session_info=data)
            self.session_info_view.update(self.session_info)
        elif "process_data" in message_type:
            self.advanced_process_data["process_data"] = data
        elif "set_sensors" in message_type:
            self.set_sensors(data)
        elif "new_calibration" in message_type:
            should_update_calibration = (
                self.current_module_info
                and self.current_module_info.calibration_mapper is not None
                and self.calibration is None
            )
            if should_update_calibration:
                self.calibration = data
                calibration_config = self.module_label_to_calibration_config_map[
                    self.current_module_label
                ]
                self.current_module_info.calibration_mapper.update_config_from_calibration(
                    configuration=calibration_config,
                    calibration=self.calibration,
                )
                self.calibration_ui_state.buffer("Session")

                if self.calibration_ui_state.auto_apply:
                    self.apply_current_calibration()

                self.refresh_calibration_pidgets()
        else:
            print("Thread data not implemented!")
            print(message_type, message, data)

    def update_external_plots(self, data):
        self.plot_queue.append(data)

    def plot_timer_fun(self):
        if not self.plot_queue:
            return

        data, *self.plot_queue = self.plot_queue[-2:]
        self.service_widget.update(data)

    def update_sweep_info(self, infos):
        if not isinstance(infos, list):  # If squeezed
            infos = [infos]

        missed = any([e.get("missed_data", False) for e in infos])
        saturated = any([e.get("data_saturated", False) for e in infos])
        data_quality_warning = any([e.get("data_quality_warning", False) for e in infos])

        if missed:
            self.num_missed_frames += 1

        self.num_recv_frames += 1

        show_lim = int(1e6)
        num_missed_show = min(self.num_missed_frames, show_lim)
        missed_sym = ">" if num_missed_show >= show_lim else ""
        num_recv_show = min(self.num_recv_frames, show_lim)
        recv_sym = ">" if num_recv_show >= show_lim else ""

        text = "Frames: {:s}{:d} (missed {:s}{:d})".format(
            recv_sym,
            num_recv_show,
            missed_sym,
            num_missed_show,
        )
        self.labels["sweep_info"].setText(text)

        tick_info = self.measured_update_rate_fc.tick_values()
        if tick_info is not None:
            _, f, _ = tick_info

            self.labels["measured_update_rate"].setText(f"{f:>10.1f} Hz")

        RED_TEXT_TIMEOUT = 2
        now = time.time()

        if missed:
            self.labels["sweep_info"].setStyleSheet("QLabel {color: red}")
            self.reset_missed_frame_text_time = now + RED_TEXT_TIMEOUT
        if self.reset_missed_frame_text_time is None or self.reset_missed_frame_text_time < now:
            self.labels["sweep_info"].setStyleSheet("")

        if data_quality_warning:
            self.labels["data_warnings"].setText("Warning: Bad data quality, restart service!")
        elif saturated:
            self.labels["data_warnings"].setText("Warning: Data saturated, reduce gain!")

        self.labels["data_warnings"].setVisible(saturated or data_quality_warning)

        if self.get_gui_state("load_state") != LoadState.LOADED:
            try:
                text = str(min(self.num_recv_frames, int(self.textboxes["sweep_buffer"].text())))
            except Exception:
                text = ""
            self.textboxes["stored_frames"].setText(text)

    def start_up(self):
        if self.under_test:
            return

        if self.use_last_config and os.path.isfile(LAST_CONF_FILENAME):
            try:
                last = np.load(LAST_CONF_FILENAME, allow_pickle=True)
                self.load_last_config(last.item())
                log.info(f"Loaded configuration from last session: {LAST_CONF_FILENAME}")
            except Exception as e:
                print("Could not load settings from last session\n{}".format(e))

    def load_last_config(self, last_config):
        # Restore sensor configs (configbase)
        dumps = last_config.get("sensor_config_dumps", {})
        for key, conf in self.module_label_to_sensor_config_map.items():
            if key in dumps:
                dump = last_config["sensor_config_dumps"][key]
                try:
                    conf._loads(dump)
                except Exception:
                    print("Could not load sensor config for '{}'".format(key))
                    conf._reset()  # TODO: load module defaults

        # Restore processing configs (configbase)
        dumps = last_config.get("processing_config_dumps", {})
        for key, conf in self.module_label_to_processing_config_map.items():
            if key in dumps:
                dump = last_config["processing_config_dumps"][key]
                try:
                    conf._loads(dump)
                except Exception:
                    print("Could not load processing config for '{}'".format(key))
                    conf._reset()

        # Restore misc. settings
        self.textboxes["sweep_buffer"].setText(last_config["sweep_buffer"])
        self.ports_dd.setCurrentIndex(last_config["port"])
        self.textboxes["host"].setText(last_config["host"])

        interface_index = last_config["interface"]
        if interface_index not in range(0, self.interface_dd.count()):
            interface_index = 0
        self.interface_dd.setCurrentIndex(interface_index)

        if last_config.get("override_baudrate"):
            self.override_baudrate = last_config["override_baudrate"]

    def closeEvent(self, event=None):
        sensor_config_dumps = {}
        for module_label, config in self.module_label_to_sensor_config_map.items():
            try:
                sensor_config_dumps[module_label] = config._dumps()
            except AttributeError:
                pass

        processing_config_dumps = {}
        for module_label, config in self.module_label_to_processing_config_map.items():
            try:
                processing_config_dumps[module_label] = config._dumps()
            except AttributeError:
                pass

        last_config = {
            "sensor_config_dumps": sensor_config_dumps,
            "processing_config_dumps": processing_config_dumps,
            "host": self.textboxes["host"].text(),
            "sweep_buffer": self.textboxes["sweep_buffer"].text(),
            "interface": self.interface_dd.currentIndex(),
            "port": self.ports_dd.currentIndex(),
            "override_baudrate": self.override_baudrate,
        }

        if self.use_last_config and not self.under_test:
            np.save(LAST_CONF_FILENAME, last_config, allow_pickle=True)
            log.info(f"Saved configuration from this session to {LAST_CONF_FILENAME}")

        try:
            self.client.disconnect()
        except Exception:
            pass

        self.close()

    def get_sensor_config(self):
        module_info = self.current_module_info

        if module_info is None:
            return None

        module_label = module_info.label
        config = self.module_label_to_sensor_config_map[module_label]

        if len(self.get_sensors()):
            config.sensor = self.get_sensors()
        else:
            config.sensor = [1]
            self.set_sensors([1])

        return config

    def get_processing_config(self, module_label=None):
        if module_label is None:
            module_info = self.current_module_info
        else:
            module_info = MODULE_LABEL_TO_MODULE_INFO_MAP[module_label]

        if module_info is None:
            return None

        module_label = module_info.label
        return self.module_label_to_processing_config_map[module_label]

    def get_default_processing_config(self, module_label=None):
        if module_label is not None:
            module_info = MODULE_LABEL_TO_MODULE_INFO_MAP[module_label]
        else:
            module_info = self.current_module_info

        if module_info is None:
            return {}

        if module_info.processing_config_class is None:
            return {}

        return module_info.processing_config_class()

    def get_calibration_config(self):
        module_info = self.current_module_info

        if module_info is None:
            return None

        return self.module_label_to_calibration_config_map.get(module_info.label)

    @property
    def in_supported_mode(self):
        try:
            return self.get_sensor_config().mode in self.client.supported_modes
        except (AttributeError, TypeError):
            return None


class Threaded_Scan(QtCore.QThread):
    sig_scan = Signal(str, str, object)

    def __init__(self, params, parent=None):
        QtCore.QThread.__init__(self, parent)

        self.client = parent.client
        self.radar = parent.radar
        self.sensor_config = params["sensor_config"]
        self.params = params
        self.data = parent.data
        self.parent = parent
        self.running = True

        self.finished.connect(self.stop_thread)

    def stop_thread(self):
        self.quit()

    def run(self):
        if self.params["data_source"] == "stream":
            record = None

            try:
                session_info = self.client.setup_session(self.sensor_config)
                self._emit("session_info", "", session_info)
                self.radar.prepare_processing(self, self.params, session_info)
                self.client.start_session()
            except et.a111.SessionSetupError:
                self.running = False
                self._emit("session_setup_error", "")
            except Exception as e:
                traceback.print_exc()
                self._emit(
                    "client_error",
                    "Failed to setup streaming!\n" "{}".format(self.format_error(e)),
                )
                self.running = False

            try:
                while self.running:
                    info, sweep = self.client.get_next()
                    self._emit("sweep_info", "", info)
                    process_results, record = self.radar.process(sweep, info)

                    if isinstance(process_results, dict) and "new_calibration" in process_results:
                        self._emit("new_calibration", "", process_results["new_calibration"])
            except Exception as e:
                traceback.print_exc()
                msg = "Failed to communicate with server!\n{}".format(self.format_error(e))
                self._emit("client_error", msg)

            try:
                self.client.stop_session()
            except Exception:
                pass

            if record is not None and len(record.data) > 0:
                self._emit("scan_data", "", record)
        elif self.params["data_source"] == "file":
            self._emit("session_info", "ok", self.data.session_info)

            try:
                self.radar.prepare_processing(self, self.params, self.data.session_info)
                self.radar.process_saved_data(self.data, self)
            except Exception as e:
                traceback.print_exc()
                error = self.format_error(e)
                self._emit("processing_error", "Error while replaying data:<br>" + error)
        else:
            self._emit("error", "Unknown mode %s!" % self.mode)
        self._emit("scan_done", "", "")

    def receive(self, message_type, message, data=None):
        if message_type == "stop":
            if self.running:
                self.running = False
                self.radar.abort_processing()
        elif message_type == "update_feature_extraction":
            self.radar.update_feature_extraction(message, data)
        elif message_type == "update_feature_list":
            self.radar.update_feature_list(data)
        elif message_type == "start":
            pass
        else:
            print("Scan thread received unknown signal: {}".format(message_type))

    def _emit(self, message_type, message, data=None):
        self.sig_scan.emit(message_type, message, data)

    def format_error(self, e):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        err = "{}\n{}\n{}\n{}".format(exc_type, fname, exc_tb.tb_lineno, e)
        return err


def sigint_handler(gui):
    event = threading.Event()
    thread = threading.Thread(target=watchdog, args=(event,))
    thread.start()
    gui.closeEvent()
    event.set()
    thread.join()


def watchdog(event):
    flag = event.wait(1)
    if not flag:
        print("\nforcing exit...")
        os._exit(1)


def remove_user_data_files():
    """Removes all files under USER_DATA_DIR interactively"""
    user_data_dir = pathlib.Path(USER_DATA_DIR)
    if not user_data_dir.exists():
        print(f'Config folder ("{user_data_dir}") does not exists.')
        print("Nothing will be done.")
        return

    if not any(user_data_dir.iterdir()):
        print(f'There exists no files under "{user_data_dir}".')
        print("Nothing will be done.")
        return

    print("Proceeding will remove the following files:\n")
    for file in user_data_dir.iterdir():
        print(f"    * {file}")
    print()

    choice = input("Continue? [y/N] ")
    should_remove = choice.lower().startswith("y")
    if should_remove:
        for file in user_data_dir.iterdir():
            file.unlink()  # unlink <=> rm
            print(f'Removed "{file}"')
    else:
        print("Nothing was removed.")


def main():
    et.utils.config_logging(level=logging.INFO)

    # Enable warnings to be printed to the log, e.g. DeprecationWarning
    warnings.simplefilter("module")

    args = ExptoolArgumentParser().parse_args()

    if args.purge_config:
        remove_user_data_files()
        sys.exit(0)

    if args.use_last_config:
        if not os.path.exists(USER_DATA_DIR):
            log.info(f"Creating folder {USER_DATA_DIR}")

        os.makedirs(USER_DATA_DIR, exist_ok=True)

    if args.portable:
        pass

    QApplication.setHighDpiScaleFactorRoundingPolicy(
        QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough,
    )
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)
    ex = GUI(use_last_config=args.use_last_config)

    signal.signal(signal.SIGINT, lambda *_: sigint_handler(ex))

    # Makes sure the signal is caught
    timer = QtCore.QTimer()
    timer.timeout.connect(lambda: None)
    timer.start(200)

    sys.exit(app.exec())



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import time

import usb.core
from usb.util import CTRL_RECIPIENT_INTERFACE, CTRL_TYPE_CLASS


class UsbPortError(Exception):
    pass


class PyUsbComm:

    device_cache = {}

    def __init__(self):
        pass

    def iterate_devices(self):
        dev = usb.core.find(find_all=True)
        for cfg in dev:
            serial_number = None
            try:
                serial_number = cfg.serial_number
            except ValueError:
                pass
            yield (cfg.idVendor, cfg.idProduct, serial_number)

    def is_accessible(self, vid, pid):
        vid_pid_str = f"{vid:04x}:{pid:04x}"

        if vid_pid_str in self.device_cache:
            return self.device_cache[vid_pid_str]
        else:
            try:
                device = usb.core.find(idVendor=vid, idProduct=pid)
                # This will raise an USBError exception if inaccessible
                device.is_kernel_driver_active(0)
                self.device_cache[vid_pid_str] = True
            except usb.core.USBError:
                self.device_cache[vid_pid_str] = False

        return self.device_cache[vid_pid_str]


class PyUsbCdc:

    USB_CDC_CMD_SEND_BREAK = 0x23
    USB_MESSAGE_TIMEOUT = 2
    USB_MESSAGE_TIMEOUT_MS = 1000 * USB_MESSAGE_TIMEOUT
    USB_PACKET_TIMEOUT_MS = 200

    def __init__(self, vid=None, pid=None, serial=None, start=True):
        self.serial = serial
        self.vid = vid
        self.pid = pid
        self._dev = None
        self._cdc_data_out_ep = None
        self._cdc_data_in_ep = None
        self.is_open = False
        self._rxremaining = b""

        if not (vid and pid):
            raise AttributeError("Must provide vid & pid of device to connect to")

        if start:
            self.open()

    def open(self):
        def match(dev):
            if self.vid == dev.idVendor and self.pid == dev.idProduct:
                if self.serial is not None:
                    try:
                        serial_number = dev.serial_number
                    except ValueError:
                        pass
                    if serial_number and serial_number == self.serial:
                        return True
                else:
                    return True
            return False

        self._dev = usb.core.find(custom_match=match)

        # Detach kernel driver
        try:
            if self._dev.is_kernel_driver_active(0):
                self._dev.detach_kernel_driver(0)
        except usb.core.USBError:
            raise UsbPortError("Could not access USB device")

        # The XC120 USB device only has one config
        config = self._dev[0]

        # Interface 1 is the CDC Data interface
        iface1 = config.interfaces()[1]
        # Interface 0 has one Endpoint for data out and one Endpoint for data in
        self._cdc_data_out_ep = iface1.endpoints()[0]
        self._cdc_data_in_ep = iface1.endpoints()[1]

        self.is_open = True

        return True

    def read(self, size=None):
        if not self.is_open:
            raise UsbPortError("Port is not open")

        rx = [self._rxremaining]
        length = len(self._rxremaining)
        self._rxremaining = b""
        end_timeout = time.monotonic() + self.USB_MESSAGE_TIMEOUT
        if size:
            while length < size:
                try:
                    c = self._dev.read(
                        self._cdc_data_in_ep.bEndpointAddress,
                        self._cdc_data_in_ep.wMaxPacketSize,
                        timeout=self.USB_PACKET_TIMEOUT_MS,
                    )
                except usb.core.USBTimeoutError:
                    pass
                if c is not None and len(c):
                    rx.append(c)
                    length += len(c)
                if time.monotonic() > end_timeout:
                    break
        else:
            c = None
            try:
                c = self._dev.read(
                    self._cdc_data_in_ep.bEndpointAddress,
                    self._cdc_data_in_ep.wMaxPacketSize,
                    timeout=self.USB_MESSAGE_TIMEOUT_MS,
                )
            except usb.core.USBTimeoutError:
                pass
            if c is not None and len(c):
                rx.append(c)
        chunk = b"".join(rx)
        if size and len(chunk) >= size:
            if self._rxremaining:
                self._rxremaining = chunk[size:] + self._rxremaining
            else:
                self._rxremaining = chunk[size:]
            chunk = chunk[0:size]
        return chunk

    def write(self, data):
        if not self.is_open:
            raise UsbPortError("Port is not open")
        self._dev.write(self._cdc_data_out_ep, data)

    def reset_input_buffer(self):
        self._rxremaining = b""
        while True:
            try:
                self._dev.read(
                    self._cdc_data_in_ep.bEndpointAddress,
                    self._cdc_data_in_ep.wMaxPacketSize,
                    timeout=self.USB_PACKET_TIMEOUT_MS,
                )
            except usb.core.USBTimeoutError:
                break

    def send_break(self, duration=0.25):
        if not self.is_open:
            raise UsbPortError("Port is not open")
        self._dev.ctrl_transfer(
            CTRL_TYPE_CLASS | CTRL_RECIPIENT_INTERFACE,
            self.USB_CDC_CMD_SEND_BREAK,
            wValue=0xFFFF,
            wIndex=0,
        )

    def disconnect(self):
        if self.is_open:
            try:
                usb.util.dispose_resources(self._dev)
                self._dev.attach_kernel_driver(0)
            except usb.core.USBError:
                pass
            self._dev = None
            self.is_open = False

    def __del__(self):
        self.disconnect()

    def close(self):
        self.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QHBoxLayout, QWidget

from . import pidgets


class TwoSensorIdsEditor(QWidget):

    sig_update = Signal(object)
    sensor_ids: t.Optional[list[int]]

    def __init__(self, name_label_texts: list[str]):
        super().__init__()

        self._sensor_id_pidget_1 = pidgets.SensorIdParameterWidgetFactory(
            name_label_text=name_label_texts[0], items=[]
        ).create(self)
        self._sensor_id_pidget_2 = pidgets.SensorIdParameterWidgetFactory(
            name_label_text=name_label_texts[1], items=[]
        ).create(self)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().addWidget(self._sensor_id_pidget_1)
        self.layout().addWidget(self._sensor_id_pidget_2)

        self._sensor_id_pidget_1.sig_parameter_changed.connect(
            lambda sensor_id: self.handle_pidget_signal(sensor_id, sensor_id_position=0)
        )
        self._sensor_id_pidget_2.sig_parameter_changed.connect(
            lambda sensor_id: self.handle_pidget_signal(sensor_id, sensor_id_position=1)
        )
        self.sensor_ids = None

    def set_selected_sensors(
        self, sensor_ids: t.Optional[list[int]], sensor_list: list[int]
    ) -> None:
        if sensor_ids:
            assert len(sensor_ids) == 2
            self.sensor_ids = sensor_ids
            self._sensor_id_pidget_1.set_selected_sensor(sensor_ids[0], sensor_list)
            self._sensor_id_pidget_2.set_selected_sensor(sensor_ids[1], sensor_list)

    def handle_pidget_signal(self, sensor_id: int, sensor_id_position: int) -> None:
        if self.sensor_ids is not None:
            self.sensor_ids[sensor_id_position] = sensor_id
            self.sig_update.emit(self.sensor_ids)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional, Type, Union

import acconeer.exptool as et
from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import ClientBase, Recorder
from acconeer.exptool.a121._rate_calc import _RateStats

from .communication_protocol import CommunicationProtocol
from .exploration_client import ExplorationClient
from .mock_client import MockClient
from .utils import get_one_usb_device


class Client(ClientBase):
    _protocol_overridden: bool
    _real_client: ClientBase

    def __init__(
        self,
        ip_address: Optional[str] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool, et.utils.USBDevice]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        _override_protocol: Optional[Type[CommunicationProtocol]] = None,
    ):
        if len([e for e in [ip_address, serial_port, usb_device, mock] if e is not None]) > 1:
            raise ValueError("Only one connection can be selected")

        if mock is not None:
            if mock:
                client_info = ClientInfo(
                    mock=mock,
                )
                self._real_client = MockClient(client_info=client_info)
            else:
                raise ValueError("mock=False is not valid")
        else:
            if isinstance(usb_device, str):
                usb_device = et.utils.get_usb_device_by_serial(usb_device, only_accessible=False)

            if isinstance(usb_device, bool):
                if usb_device:
                    usb_device = get_one_usb_device()
                else:
                    raise ValueError("usb_device=False is not valid")

            client_info = ClientInfo(
                ip_address=ip_address,
                override_baudrate=override_baudrate,
                serial_port=serial_port,
                usb_device=usb_device,
            )

            self._real_client = ExplorationClient(
                client_info=client_info,
                _override_protocol=_override_protocol,
            )

    def connect(self) -> None:
        self._real_client.connect()

    def __enter__(self) -> Client:
        self._real_client.connect()
        return self

    def __exit__(self, *_: Any) -> None:
        self._real_client.disconnect()

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        return self._real_client.setup_session(config, calibrations)

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        self._real_client.start_session(recorder)

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        return self._real_client.get_next()

    def stop_session(self) -> Any:
        return self._real_client.stop_session()

    def disconnect(self) -> None:
        self._real_client.disconnect()

    @property
    def connected(self) -> bool:
        return self._real_client.connected

    @property
    def session_is_setup(self) -> bool:
        return self._real_client.session_is_setup

    @property
    def session_is_started(self) -> bool:
        return self._real_client.session_is_started

    @property
    def server_info(self) -> ServerInfo:
        return self._real_client.server_info

    @property
    def client_info(self) -> ClientInfo:
        return self._real_client.client_info

    @property
    def session_config(self) -> SessionConfig:
        return self._real_client.session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        return self._real_client.extended_metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        return self._real_client.calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        return self._real_client.calibrations_provided

    @property
    def _rate_stats(self) -> _RateStats:
        return self._real_client._rate_stats



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from acconeer.exptool import a121

from ._detector import DetectorConfig


def get_high_accuracy_detector_config() -> DetectorConfig:
    return DetectorConfig(
        max_step_length=2, signal_quality=20.0, max_profile=a121.Profile.PROFILE_3
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc

import pyqtgraph as pg

from acconeer.exptool.app.new.app_model import AppModel, PlotPluginInterface
from acconeer.exptool.app.new.backend import GeneralMessage

from .ui_plugin_base import UiPluginBase


class PlotPluginBase(UiPluginBase, abc.ABC, PlotPluginInterface):
    def __init__(self, app_model: AppModel, plot_layout: pg.GraphicsLayout) -> None:
        super().__init__(app_model)

        self.__app_model = app_model
        self.__app_model.sig_message_plot_plugin.connect(self.handle_message)

        self.plot_layout = plot_layout

    def stop_listening(self) -> None:
        super().stop_listening()
        self.__app_model.sig_message_plot_plugin.disconnect(self.handle_message)

    @abc.abstractmethod
    def handle_message(self, message: GeneralMessage) -> None:
        pass

    @abc.abstractmethod
    def draw(self) -> None:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import tempfile

import h5py
import numpy as np

import acconeer.exptool as et


class ProcessingTestModule:
    def __init__(self, processor_class, processing_config_class, path, test_keys, parameter_sets):

        self.processor_class = processor_class
        self.processsing_config_class = processing_config_class
        self.path = path
        self.test_keys = test_keys
        self.parameter_sets = parameter_sets

    def get_output(self, parameter_set=None):
        input_record = et.a111.recording.load(self.path / "input.h5")

        processing_config = self.processsing_config_class()

        if parameter_set is not None:
            for k, v in parameter_set.items():
                setattr(processing_config, k, v)

        processor = self.processor_class(
            input_record.sensor_config,
            processing_config,
            input_record.session_info,
        )

        output = {k: [] for k in self.test_keys}

        for data_info, data in input_record:
            if data.shape[0] == 1:
                result = processor.process(data.squeeze(0), data_info[0])
            else:
                result = processor.process(data, data_info[0])
            if result is not None:

                for k in self.test_keys:
                    output[k].append(result[k])

        # Explicit `dtype=float` makes the conversion `None` -> `np.nan`.
        return {k: np.array(v, dtype=float) for k, v in output.items()}

    def save_output(self, file, output):
        with h5py.File(file, "w") as f:
            for k in self.test_keys:
                try:
                    f.create_dataset(name=k, data=output[k], track_times=False, compression="gzip")
                except TypeError as te:
                    raise TypeError(
                        f"Could not create dataset with name: {k}, data={output[k]}"
                    ) from te

    def load_output(self, file):
        output = {}

        with h5py.File(file, "r") as f:
            for k in self.test_keys:
                output[k] = f[k][()]

        return output

    def compare_output(self, expected, actual, exact=False):
        for k in self.test_keys:
            expected_arr = expected[k]
            actual_arr = actual[k]

            if exact:
                assert np.all(np.array_equal(expected_arr, actual_arr, equal_nan=True))
            else:
                assert np.all(np.isclose(expected_arr, actual_arr, equal_nan=True))

    def path_for_parameter_set(self, parameter_set):
        if parameter_set:
            l = sorted(parameter_set.items())
            suffix = "_".join(f"{k}-{v}" for k, v in l)
        else:
            suffix = "default"

        return self.path / f"output_{suffix}.h5"

    def test_path_for_parameter_set(self):
        assert self.path_for_parameter_set({"foo": "bar"}) == (self.path / "output_foo-bar.h5")

    def test_load_save_compare(self):
        temp_file = tempfile.TemporaryFile()

        saved_output = self.get_output()

        self.save_output(temp_file, saved_output)
        loaded_output = self.load_output(temp_file)

        self.compare_output(saved_output, loaded_output, exact=True)

    def test_processor_against_reference(self):
        for parameter_set in self.parameter_sets:
            with open(self.path_for_parameter_set(parameter_set), "rb") as f:
                expected = self.load_output(f)

            actual = self.get_output(parameter_set)
            self.compare_output(expected, actual)

    def run_all_tests(self):
        self.test_load_save_compare()
        self.test_processor_against_reference()
        self.test_path_for_parameter_set()

    def main(self):
        import argparse

        parser = argparse.ArgumentParser()

        subparsers = parser.add_subparsers(dest="command")
        subparsers.required = True

        subparsers.add_parser("save")

        args = parser.parse_args()

        if args.command == "save":
            for parameter_set in self.parameter_sets:
                output = self.get_output(parameter_set)
                self.save_output(self.path_for_parameter_set(parameter_set), output)
        else:
            raise RuntimeError



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import logging
from functools import partial
from typing import Any, Mapping, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QTabWidget, QToolButton, QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from . import pidgets
from .subsweep_config_editor import SubsweepConfigEditor
from .types import PidgetFactoryMapping
from .utils import VerticalGroupBox


log = logging.getLogger(__name__)


class SensorConfigEditor(QWidget):
    sig_update = Signal(object)

    _sensor_config: Optional[a121.SensorConfig]
    _subsweep_config_editors: list[SubsweepConfigEditor]

    _all_pidgets: list[pidgets.ParameterWidget]

    _read_only: bool
    _supports_multiple_subsweeps: bool

    SPACING = 15
    IDLE_STATE_LABEL_MAP = {
        a121.IdleState.READY: "Ready",
        a121.IdleState.SLEEP: "Sleep",
        a121.IdleState.DEEP_SLEEP: "Deep sleep",
    }
    SENSOR_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "sweeps_per_frame": pidgets.IntParameterWidgetFactory(
            name_label_text="Sweeps per frame:",
            name_label_tooltip=(
                "The number of sweeps that will be captured in each frame (measurement)."
            ),
            limits=(1, 4095),
        ),
        "sweep_rate": pidgets.OptionalFloatParameterWidgetFactory(
            name_label_text="Sweep rate:",
            name_label_tooltip=(
                "The sweep rate for sweeps in a frame (measurement).\n"
                "If 'Limit' is unchecked, the sweep rate will be as fast as possible."
            ),
            limits=(1, 1e6),
            decimals=0,
            init_set_value=1000.0,
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "frame_rate": pidgets.OptionalFloatParameterWidgetFactory(
            name_label_text="Frame rate:",
            name_label_tooltip=(
                "Frame rate.\nIf 'Limit' is unchecked, the rate is not limited by the sensor "
                "but by the rate that the server acknowledge and reads out the frame."
            ),
            limits=(0.1, 1e4),
            decimals=1,
            init_set_value=10.0,
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "inter_sweep_idle_state": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter sweep idle state:",
            name_label_tooltip=(
                "The inter sweep idle state is the state the sensor idles in "
                "between each sweep in a frame."
            ),
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "inter_frame_idle_state": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter frame idle state:",
            name_label_tooltip=(
                "The inter frame idle state is the state the sensor idles in between each frame."
            ),
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "continuous_sweep_mode": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Continuous sweep mode",
            name_label_tooltip=(
                "<p>With CSM, the sensor timing is set up to generate a continuous "
                "stream of sweeps, even if more than one sweep per frame is used. "
                "The interval between the last sweep in one frame to the first "
                "sweep in the next frame becomes equal to the interval between "
                "sweeps within a frame (given by the sweep rate).</p>"
                ""
                "<p>It ensures that:</p>"
                ""
                "<p>'frame rate' = 'sweep rate' / 'sweeps per frame'</p>"
                ""
                "While the frame rate parameter can be set to approximately "
                "satisfy this condition, using CSM is more precise."
                ""
                "<p>If only one sweep per frame is used, CSM has no use since a "
                "continuous stream of sweeps is already given (if a fixed frame "
                "rate is used).</p>"
                ""
                "<p>The main use for CSM is to allow reading out data at a slower "
                "rate than the sweep rate, while maintaining that sweep rate "
                "continuously.</p>"
                ""
                "<p>Note that in most cases, double buffering must be enabled to "
                "allow high rates without delays.</p>"
            ),
        ),
        "double_buffering": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Double buffering",
            name_label_tooltip=(
                "Double buffering will split the sensor buffer in two halves. "
                "One half is used to read out the frame while sampling is done "
                "in the other half."
            ),
        ),
    }

    def __init__(
        self, supports_multiple_subsweeps: bool = False, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)

        self._all_pidgets = []

        self._sensor_config = None
        self._subsweep_config_editors = []

        self._read_only = False
        self._supports_multiple_subsweeps = supports_multiple_subsweeps

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.sensor_group_box = VerticalGroupBox("Sensor parameters", parent=self)
        self.sensor_group_box.layout().setSpacing(self.SPACING)
        self.layout().addWidget(self.sensor_group_box)

        self._sensor_config_pidgets: Mapping[str, pidgets.ParameterWidget] = {}
        for aspect, factory in self.SENSOR_CONFIG_FACTORIES.items():
            pidget = factory.create(self.sensor_group_box)
            self.sensor_group_box.layout().addWidget(pidget)

            pidget.sig_parameter_changed.connect(
                partial(self._update_sensor_config_aspect, aspect)
            )

            self._all_pidgets.append(pidget)
            self._sensor_config_pidgets[aspect] = pidget

        self.subsweep_group_box = VerticalGroupBox("Subsweep parameters", parent=self)
        self.subsweep_group_box.layout().setSpacing(self.SPACING)
        self.layout().addWidget(self.subsweep_group_box)

        self._tab_widget = QTabWidget(self)
        self._tab_widget.setStyleSheet("QTabWidget::pane { padding: 5px;}")
        self.subsweep_group_box.layout().addWidget(self._tab_widget)

        if supports_multiple_subsweeps:
            self._tab_widget.setTabsClosable(True)
            self._tab_widget.tabCloseRequested.connect(self._remove_subsweep_config)

            self._plus_button = QToolButton(self)
            self._plus_button.setText("+")
            self.layout().addWidget(self._plus_button)
            self._plus_button.clicked.connect(self._add_subsweep_config)
            self._tab_widget.setCornerWidget(self._plus_button)
            self._tab_widget.cornerWidget().setMinimumSize(self._plus_button.sizeHint())

    def _add_subsweep_config(self) -> None:
        if self._sensor_config is None:
            return
        if self._tab_widget.count() > 3:
            return
        subsweep_config = a121.SubsweepConfig()
        self._sensor_config._subsweeps.append(subsweep_config)
        self._broadcast()
        subsweep_config_editor = self._add_subsweep_config_editor()
        subsweep_config_editor.set_data(subsweep_config)
        subsweep_config_editor.sync()

    def _add_tabs(self, tabs_needed: int) -> None:
        while self._tab_widget.count() < tabs_needed:
            self._add_subsweep_config_editor()

    def _add_subsweep_config_editor(self) -> SubsweepConfigEditor:
        subsweep_config_editor = SubsweepConfigEditor(self)
        subsweep_config_editor.sig_update.connect(self._broadcast)
        subsweep_config_editor.set_read_only(self._read_only)
        self._subsweep_config_editors.append(subsweep_config_editor)
        self._tab_widget.addTab(subsweep_config_editor, str(len(self._subsweep_config_editors)))
        self._update_tab_labels()
        return subsweep_config_editor

    def _remove_subsweep_config(self, idx: int) -> None:
        if self._sensor_config is None:
            return
        if self._tab_widget.count() < 2:
            return
        self._sensor_config.subsweeps.pop(idx)
        self._broadcast()
        self._remove_subsweep_config_editor(idx)

    def _remove_tabs(self, tabs_needed: int) -> None:
        while self._tab_widget.count() > tabs_needed:
            self._remove_subsweep_config_editor(0)

    def _remove_subsweep_config_editor(self, idx: int) -> None:
        self._tab_widget.removeTab(idx)
        self._subsweep_config_editors.pop(idx)
        self._update_tab_labels()

    def _update_tab_labels(self) -> None:
        for tab_idx in range(self._tab_widget.count()):
            self._tab_widget.setTabText(tab_idx, f"{tab_idx + 1}      ")

    def _update_ui(self) -> None:
        if self._sensor_config is None:
            log.debug("could not update ui as SensorConfig is None")
            return

        self._sensor_config_pidgets["sweeps_per_frame"].set_parameter(
            self._sensor_config.sweeps_per_frame
        )
        self._sensor_config_pidgets["sweep_rate"].set_parameter(self._sensor_config.sweep_rate)
        self._sensor_config_pidgets["frame_rate"].set_parameter(self._sensor_config.frame_rate)
        self._sensor_config_pidgets["continuous_sweep_mode"].set_parameter(
            self._sensor_config.continuous_sweep_mode
        )
        self._sensor_config_pidgets["double_buffering"].set_parameter(
            self._sensor_config.double_buffering
        )
        self._sensor_config_pidgets["inter_frame_idle_state"].set_parameter(
            self._sensor_config.inter_frame_idle_state
        )
        self._sensor_config_pidgets["inter_sweep_idle_state"].set_parameter(
            self._sensor_config.inter_sweep_idle_state
        )

    def set_data(self, sensor_config: Optional[a121.SensorConfig]) -> None:
        self._sensor_config = sensor_config
        if sensor_config is None:
            return
        tabs_needed = len(sensor_config.subsweeps)
        self._remove_tabs(tabs_needed)
        self._add_tabs(tabs_needed)
        for i, subsweep in enumerate(sensor_config.subsweeps):
            self._subsweep_config_editors[i].set_data(subsweep)

    def set_read_only(self, read_only: bool) -> None:
        self._read_only = read_only

        for pidget in self._all_pidgets:
            pidget.setEnabled(not read_only)

        if self._supports_multiple_subsweeps:
            self._tab_widget.setTabsClosable(not read_only)
            self._plus_button.setEnabled(not read_only)

        for editor in self._subsweep_config_editors:
            editor.set_read_only(read_only)

    def sync(self) -> None:
        self._update_ui()
        for subsweep_config_editor in self._subsweep_config_editors:
            subsweep_config_editor.sync()

    def _broadcast(self) -> None:
        self.sig_update.emit(self._sensor_config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        for pidget in self._all_pidgets:
            pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        for subsweep_config_editor in self._subsweep_config_editors:
            unhandled_results = subsweep_config_editor.handle_validation_results(unhandled_results)

        return unhandled_results

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None or self._sensor_config is None:
            return False

        result_handled = False

        if result.source == self._sensor_config:
            for pidget in self._sensor_config_pidgets:
                if result.aspect == pidget:
                    self._sensor_config_pidgets[result.aspect].set_note_text(
                        result.message, result.criticality
                    )
                    result_handled = True

        return result_handled

    def _update_sensor_config_aspect(self, aspect: str, value: Any) -> None:
        if self._sensor_config is None:
            raise TypeError("SensorConfig is None")

        try:
            setattr(self._sensor_config, aspect, value)
        except Exception as e:
            self._sensor_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)

        self._broadcast()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional, Union

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import ClientBase, ClientError, Recorder
from acconeer.exptool.a121._core.utils import unextend
from acconeer.exptool.a121._rate_calc import _RateCalculator, _RateStats


class CommonClient(ClientBase):
    _client_info: ClientInfo
    _calibrations_provided: dict[int, bool]
    _metadata: Optional[list[dict[int, Metadata]]]
    _rate_stats_calc: Optional[_RateCalculator]
    _recorder: Optional[Recorder]
    _sensor_calibrations: Optional[dict[int, SensorCalibration]]
    _session_config: Optional[SessionConfig]
    _session_is_started: bool

    def _assert_connected(self) -> None:
        if not self.connected:
            raise ClientError("Client is not connected.")

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            raise ClientError("Session is not started.")

    def __init__(self, client_info: ClientInfo) -> None:
        self._client_info = client_info
        self._calibrations_provided = {}
        self._metadata = None
        self._rate_stats_calc = None
        self._recorder = None
        self._sensor_calibrations = None
        self._session_is_started = False
        self._session_config = None

    def _recorder_start(self, recorder: Optional[Recorder]) -> None:
        if recorder is not None:
            calibrations_provided: Optional[dict[int, bool]] = self.calibrations_provided
            try:
                calibrations = self.calibrations
            except ClientError:
                calibrations = None
                calibrations_provided = None

            self._recorder = recorder
            self._recorder._start(
                client_info=self.client_info,
                extended_metadata=self.extended_metadata,
                server_info=self.server_info,
                session_config=self.session_config,
                calibrations=calibrations,
                calibrations_provided=calibrations_provided,
            )

    def _recorder_stop(self) -> Any:
        if self._recorder is None:
            recorder_result = None
        else:
            recorder_result = self._recorder._stop()
            self._recorder = None
        return recorder_result

    def _recorder_sample(self, extended_results: list[dict[int, Result]]) -> None:
        if self._recorder is not None:
            self._recorder._sample(extended_results)

    def _create_rate_stats_calc(self) -> None:
        assert self._metadata is not None
        self._rate_stats_calc = _RateCalculator(self.session_config, self._metadata)

    def _update_rate_stats_calc(self, extended_results: list[dict[int, Result]]) -> None:
        assert self._rate_stats_calc is not None
        self._rate_stats_calc.update(extended_results)

    def _return_results(
        self, extended_results: list[dict[int, Result]]
    ) -> Union[Result, list[dict[int, Result]]]:
        if self.session_config.extended:
            return extended_results
        else:
            return unextend(extended_results)

    @property
    def session_is_setup(self) -> bool:
        return self._metadata is not None

    @property
    def session_is_started(self) -> bool:
        return self._session_is_started

    @property
    def client_info(self) -> ClientInfo:
        return self._client_info

    @property
    def session_config(self) -> SessionConfig:
        self._assert_session_setup()
        assert self._session_config is not None  # Should never happen if session is setup
        return self._session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        self._assert_session_setup()
        assert self._metadata is not None  # Should never happen if session is setup
        return self._metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        self._assert_session_setup()

        if not self._sensor_calibrations:
            raise ClientError("Server did not provide calibration")

        return self._sensor_calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        return self._calibrations_provided

    @property
    def _rate_stats(self) -> _RateStats:
        self._assert_session_started()
        assert self._rate_stats_calc is not None
        return self._rate_stats_calc.stats



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.history_length_s = processing_config.history_length_s
        self.depths = et.a111.get_range_depths(sensor_config, session_info)

        max_num_of_sectors = max(6, self.depths.size // 3)
        self.sector_size = max(1, -(-self.depths.size // max_num_of_sectors))
        self.num_sectors = -(-self.depths.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.depths.size) // 2

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer presence detection example")

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.DashLine)

        # Data plot

        self.data_plot = win.addPlot(
            row=0,
            col=0,
            title="Frame (blue), fast (orange), and slow (green)",
        )
        self.data_plot.setMenuEnabled(False)
        self.data_plot.setMouseEnabled(x=False, y=False)
        self.data_plot.hideButtons()
        self.data_plot.showGrid(x=True, y=True)
        self.data_plot.setLabel("bottom", "Depth (m)")
        self.data_plot.setLabel("left", "Amplitude")
        self.data_plot.setYRange(0, 2**16)
        self.frame_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.fast_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.slow_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(2),
        )
        self.data_plot.addItem(self.frame_scatter)
        self.data_plot.addItem(self.fast_scatter)
        self.data_plot.addItem(self.slow_scatter)
        self.frame_smooth_limits = et.utils.SmoothLimits(self.sensor_config.update_rate)

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=1,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Depth (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)

        # Depthwise presence plot

        self.move_plot = win.addPlot(
            row=2,
            col=0,
            title="Depthwise presence",
        )
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Depth (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        zero_curve = self.move_plot.plot(self.depths, np.zeros_like(self.depths))
        self.inter_curve = self.move_plot.plot()
        self.total_curve = self.move_plot.plot()
        self.move_smooth_max = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        fbi = pg.FillBetweenItem(
            zero_curve,
            self.inter_curve,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.move_plot.addItem(fbi)

        fbi = pg.FillBetweenItem(
            self.inter_curve,
            self.total_curve,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.move_plot.addItem(fbi)

        # Presence history plot

        self.move_hist_plot = pg.PlotItem(title="Presence history")
        self.move_hist_plot.setMenuEnabled(False)
        self.move_hist_plot.setMouseEnabled(x=False, y=False)
        self.move_hist_plot.hideButtons()
        self.move_hist_plot.showGrid(x=True, y=True)
        self.move_hist_plot.setLabel("bottom", "Time (s)")
        self.move_hist_plot.setLabel("left", "Score")
        self.move_hist_plot.setXRange(-self.history_length_s, 0)
        self.history_smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)
        self.move_hist_plot.setYRange(0, 10)

        self.move_hist_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler())
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_hist_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_hist_plot.addItem(self.present_text_item)
        self.move_hist_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=3, col=0)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_hist_plot, col=0)
        sublayout.addItem(self.sector_plot, col=1)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.data_plot.setVisible(self.processing_config.show_data)
        self.noise_plot.setVisible(self.processing_config.show_noise)
        self.move_plot.setVisible(self.processing_config.show_depthwise_output)
        self.sector_plot.setVisible(self.processing_config.show_sectors)

        for line in self.limit_lines:
            line.setPos(processing_config.detection_threshold)

    def update(self, data):
        self.frame_scatter.setData(
            np.tile(self.depths, self.sensor_config.sweeps_per_frame),
            data["frame"].flatten(),
        )

        self.fast_scatter.setData(self.depths, data["fast"])
        self.slow_scatter.setData(self.depths, data["slow"])
        self.data_plot.setYRange(*self.frame_smooth_limits.update(data["frame"]))

        noise = data["noise"]
        self.noise_curve.setData(self.depths, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data["presence_distance"]

        move_ys = data["depthwise_presence"]
        self.inter_curve.setData(self.depths, data["inter"])
        self.total_curve.setData(self.depths, move_ys)
        m = self.move_smooth_max.update(np.max(move_ys))
        m = max(m, 2 * self.processing_config.detection_threshold)
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data["presence_detected"]))

        move_hist_ys = data["presence_history"]
        move_hist_xs = np.linspace(-self.history_length_s, 0, len(move_hist_ys))

        m_hist = max(np.max(move_hist_ys), self.processing_config.detection_threshold * 1.05)
        m_hist = self.history_smooth_max.update(m_hist)

        if self.processing_config.history_plot_ceiling is not None:
            self.move_hist_plot.setYRange(0, self.processing_config.history_plot_ceiling)
            self.move_hist_curve.setData(
                move_hist_xs,
                np.minimum(move_hist_ys, self.processing_config.history_plot_ceiling),
            )
            self.set_present_text_y_pos(self.processing_config.history_plot_ceiling)
        else:
            self.move_hist_plot.setYRange(0, m_hist)
            self.move_hist_curve.setData(move_hist_xs, move_hist_ys)
            self.set_present_text_y_pos(m_hist)

        if data["presence_detected"]:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data["presence_detected"]:
            index = (data["presence_distance_index"] + self.sector_offset) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y):
        self.present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.not_present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities import Criticality


class MiscErrorView(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

    def handle_validation_results(
        self, validation_results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        COLOR_MAP = {
            Criticality.ERROR: "#E6635A",
            Criticality.WARNING: "#FCC842",
            None: "white",
        }

        # Delete all existing error message widgets
        while self.layout().count():
            child = self.layout().takeAt(0)
            if child.widget():
                child.widget().deleteLater()

        for result in validation_results:
            error_widget = QLabel(parent=self)
            error_widget.setWordWrap(True)
            error_widget.setContentsMargins(5, 5, 5, 5)
            error_widget.setText(result.message)
            error_widget.setStyleSheet(
                f"background-color: {COLOR_MAP[result.criticality]}; "
                "color: white; font: bold italic;"
            )
            self.layout().addWidget(error_widget)

        return []



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo
from acconeer.exptool.a111.algo.utils import (
    MultiSensorPGUpdaterCreator,
    MultiSensorProcessorCreator,
)

from ._processor import ProcessingConfiguration, Processor, get_sensor_config
from .ui import PGUpdater


processor_creator = MultiSensorProcessorCreator(Processor)
pg_updater_creator = MultiSensorPGUpdaterCreator(PGUpdater)


module_info = ModuleInfo(
    key="sparse_presence",
    label="Presence detection (sparse)",
    pg_updater=pg_updater_creator.create_pg_updater,
    processing_config_class=ProcessingConfiguration,
    module_family=ModuleFamily.DETECTOR,
    sensor_config_class=get_sensor_config,
    processor=processor_creator.create_processor,
    multi_sensor=True,
    docs_url="https://docs.acconeer.com/en/latest/processing/presence_detection_sparse.html",
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from PySide6 import QtCore
from PySide6.QtWidgets import QToolButton, QVBoxLayout, QWidget


class CollapsibleWidget(QWidget):
    def __init__(self, label: str, widget: QWidget, parent: QWidget) -> None:
        super().__init__(parent)

        self.widget = widget
        self.widget.setVisible(False)

        self.arrow_button = QToolButton()
        self.arrow_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.arrow_button.setText(label)
        self.arrow_button.setArrowType(QtCore.Qt.RightArrow)
        self.arrow_button.setCheckable(True)
        self.arrow_button.setChecked(False)
        self.arrow_button.toggled.connect(self._on_toggle)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.arrow_button)
        self.layout().addWidget(self.widget)

    def _on_toggle(self, checked: bool) -> None:
        self.widget.setVisible(checked)

        if checked:
            self.arrow_button.setArrowType(QtCore.Qt.DownArrow)
        else:
            self.arrow_button.setArrowType(QtCore.Qt.RightArrow)



import ctypes
import re
from ctypes import (
    byref,
    c_byte,
    c_ubyte,
    c_ulong,
    c_void_p,
    create_string_buffer,
    resize,
    sizeof,
    wstring_at,
)
from ctypes.wintypes import DWORD

from .winusb import WinUSBApi
from .winusbclasses import (
    DIGCF_DEVICE_INTERFACE,
    DIGCF_PRESENT,
    FILE_ATTRIBUTE_NORMAL,
    FILE_FLAG_OVERLAPPED,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    GENERIC_READ,
    GENERIC_WRITE,
    GUID,
    INVALID_HANDLE_VALUE,
    OPEN_EXISTING,
    PipeInfo,
    SpDeviceInterfaceData,
    SpDeviceInterfaceDetailData,
    SpDevinfoData,
    UsbInterfaceDescriptor,
)
from .winusbutils import (
    Close_Handle,
    CreateFile,
    GetLastError,
    SetupDiEnumDeviceInterfaces,
    SetupDiGetClassDevs,
    SetupDiGetDeviceInterfaceDetail,
    WinUsb_ControlTransfer,
    WinUsb_FlushPipe,
    WinUsb_Free,
    WinUsb_GetAssociatedInterface,
    WinUsb_Initialize,
    WinUsb_QueryInterfaceSettings,
    WinUsb_QueryPipe,
    WinUsb_ReadPipe,
    WinUsb_SetPipePolicy,
    WinUsb_WritePipe,
)


class WinUsbPy(object):
    def __init__(self):
        self.api = WinUSBApi()
        byte_array = c_byte * 8
        self.usb_winusb_guid = GUID(
            0xDEE824EF, 0x729B, 0x4A0E, byte_array(0x9C, 0x14, 0xB7, 0x11, 0x7D, 0x33, 0xA8, 0x17)
        )
        self.handle_file = INVALID_HANDLE_VALUE
        self.handle_winusb = [c_void_p()]
        self._index = -1

    def find_all_devices(self):
        device_path_list = []
        self.handle = self.api.exec_function_setupapi(
            SetupDiGetClassDevs,
            byref(self.usb_winusb_guid),
            None,
            None,
            DWORD(DIGCF_PRESENT | DIGCF_DEVICE_INTERFACE),
        )
        sp_device_interface_data = SpDeviceInterfaceData()
        sp_device_interface_data.cb_size = sizeof(sp_device_interface_data)
        sp_device_interface_detail_data = SpDeviceInterfaceDetailData()
        sp_device_info_data = SpDevinfoData()
        sp_device_info_data.cb_size = sizeof(sp_device_info_data)

        i = 0
        required_size = DWORD(0)
        member_index = DWORD(i)
        cb_sizes = (8, 6, 5)  # different on 64 bit / 32 bit etc

        while self.api.exec_function_setupapi(
            SetupDiEnumDeviceInterfaces,
            self.handle,
            None,
            byref(self.usb_winusb_guid),
            member_index,
            byref(sp_device_interface_data),
        ):
            self.api.exec_function_setupapi(
                SetupDiGetDeviceInterfaceDetail,
                self.handle,
                byref(sp_device_interface_data),
                None,
                0,
                byref(required_size),
                None,
            )
            resize(sp_device_interface_detail_data, required_size.value)

            path = None
            for cb_size in cb_sizes:
                sp_device_interface_detail_data.cb_size = cb_size
                ret = self.api.exec_function_setupapi(
                    SetupDiGetDeviceInterfaceDetail,
                    self.handle,
                    byref(sp_device_interface_data),
                    byref(sp_device_interface_detail_data),
                    required_size,
                    byref(required_size),
                    byref(sp_device_info_data),
                )
                if ret:
                    cb_sizes = (cb_size,)
                    path = wstring_at(byref(sp_device_interface_detail_data, sizeof(DWORD)))
                    break
            if path is None:
                raise ctypes.WinError()

            pattern = r"(.*)#vid_(?P<vid>\w+)&pid_(?P<pid>\w+)(&mi_.*)?#(?P<serial>.*)#(.*)"
            match = re.fullmatch(pattern, path)
            groups = match.groupdict()
            vid = int(f"0x{groups['vid']}", 16)
            pid = int(f"0x{groups['pid']}", 16)
            serial_number = groups["serial"]
            if "&" in serial_number:
                serial_number = None

            device_path_list.append(
                {
                    "vid": vid,
                    "pid": pid,
                    "serial": serial_number,
                    "path": path,
                }
            )

            i += 1
            member_index = DWORD(i)
            required_size = DWORD(0)
            resize(sp_device_interface_detail_data, sizeof(SpDeviceInterfaceDetailData))

        return device_path_list

    def init_winusb_device(self, path):
        if path is None:
            return False

        self.handle_file = self.api.exec_function_kernel32(
            CreateFile,
            path,
            GENERIC_WRITE | GENERIC_READ,
            FILE_SHARE_WRITE | FILE_SHARE_READ,
            None,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
            None,
        )

        if self.handle_file == INVALID_HANDLE_VALUE:
            return False
        result = self.api.exec_function_winusb(
            WinUsb_Initialize, self.handle_file, byref(self.handle_winusb[0])
        )
        if result == 0:
            # err = self.get_last_error_code()
            raise ctypes.WinError()
            # return False
        else:
            self._index = 0
            return True

    def close_winusb_device(self):
        result_file = 1
        if self.handle_file:
            result_file = self.api.exec_function_kernel32(Close_Handle, self.handle_file)
            if result_file:
                self.handle_file = None

        result_winusb = [self.api.exec_function_winusb(WinUsb_Free, h) for h in self.handle_winusb]
        if 0 in result_winusb:
            raise RuntimeError("Unable to close winusb handle")
        self.handle_winusb = []
        return result_file != 0

    def get_last_error_code(self):
        return self.api.exec_function_kernel32(GetLastError)

    def query_interface_settings(self, index):
        if self._index != -1:
            temp_handle_winusb = self.handle_winusb[self._index]
            interface_descriptor = UsbInterfaceDescriptor()
            result = self.api.exec_function_winusb(
                WinUsb_QueryInterfaceSettings,
                temp_handle_winusb,
                c_ubyte(0),
                byref(interface_descriptor),
            )
            if result != 0:
                return interface_descriptor
            else:
                return None
        else:
            return None

    def change_interface(self, index, alternate=0):
        new_handle = c_void_p()
        result = self.api.exec_function_winusb(
            WinUsb_GetAssociatedInterface,
            self.handle_winusb[self._index],
            c_ubyte(alternate),
            byref(new_handle),
        )
        if result != 0:
            self._index = index + 1
            self.handle_winusb.append(new_handle)
            return True
        else:
            return False

    def query_pipe(self, pipe_index):
        pipe_info = PipeInfo()
        result = self.api.exec_function_winusb(
            WinUsb_QueryPipe,
            self.handle_winusb[self._index],
            c_ubyte(0),
            pipe_index,
            byref(pipe_info),
        )
        if result != 0:
            return pipe_info
        else:
            return None

    def control_transfer(self, setup_packet, buff=None):
        if buff is not None:
            if setup_packet.length > 0:  # Host 2 Device
                buff = (c_ubyte * setup_packet.length)(*buff)
                buffer_length = setup_packet.length
            else:  # Device 2 Host
                buff = (c_ubyte * setup_packet.length)()
                buffer_length = setup_packet.length
        else:
            buff = c_ubyte()
            buffer_length = 0

        result = self.api.exec_function_winusb(
            WinUsb_ControlTransfer,
            self.handle_winusb[0],
            setup_packet,
            byref(buff),
            c_ulong(buffer_length),
            byref(c_ulong(0)),
            None,
        )
        return {"result": result != 0, "buffer": [buff]}

    def write(self, pipe_id, write_buffer):
        write_buffer = create_string_buffer(write_buffer)
        written = c_ulong(0)
        self.api.exec_function_winusb(
            WinUsb_WritePipe,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            write_buffer,
            c_ulong(len(write_buffer) - 1),
            byref(written),
            None,
        )
        return written.value

    def read(self, pipe_id, length_buffer):
        read_buffer = create_string_buffer(length_buffer)
        read = c_ulong(0)
        result = self.api.exec_function_winusb(
            WinUsb_ReadPipe,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            read_buffer,
            c_ulong(length_buffer),
            byref(read),
            None,
        )
        if result != 0:
            if read.value != length_buffer:
                return read_buffer[: read.value]
            else:
                return read_buffer
        else:
            return None

    def set_timeout(self, pipe_id, timeout):
        class POLICY_TYPE:
            SHORT_PACKET_TERMINATE = 1
            AUTO_CLEAR_STALL = 2
            PIPE_TRANSFER_TIMEOUT = 3
            IGNORE_SHORT_PACKETS = 4
            ALLOW_PARTIAL_READS = 5
            AUTO_FLUSH = 6
            RAW_IO = 7

        policy_type = c_ulong(POLICY_TYPE.PIPE_TRANSFER_TIMEOUT)
        value_length = c_ulong(4)
        value = c_ulong(int(timeout * 1000))  # in ms
        result = self.api.exec_function_winusb(
            WinUsb_SetPipePolicy,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            policy_type,
            value_length,
            byref(value),
        )
        return result

    def flush(self, pipe_id):
        result = self.api.exec_function_winusb(
            WinUsb_FlushPipe, self.handle_winusb[self._index], c_ubyte(pipe_id)
        )
        return result



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from acconeer.exptool.a121._core.peripherals.communication.message import Message

from .parse_error import ParseError


class SetBaudrateResponseHeader(te.TypedDict):
    status: te.Literal["ok"]
    message: te.Literal["set baudrate"]


@attrs.frozen
class SetBaudrateResponse(Message):
    @classmethod
    def parse(cls, header: dict[str, t.Any], payload: bytes) -> SetBaudrateResponse:
        t.cast(SetBaudrateResponseHeader, header)

        if header.get("status") == "ok" and header.get("message") == "set baudrate":
            return cls()
        else:
            raise ParseError



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional

from typing_extensions import Protocol

from acconeer.exptool.a121._core.entities import SensorCalibration, SessionConfig

from .message import Message


class CommunicationProtocol(Protocol):
    end_sequence: bytes

    @classmethod
    def get_system_info_command(cls) -> bytes:
        """The `get_system_info` command."""
        ...

    @classmethod
    def get_sensor_info_command(cls) -> bytes:
        """The `get_sensor_info` command."""
        ...

    @classmethod
    def setup_command(
        cls,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> bytes:
        """The `setup` command."""
        ...

    @classmethod
    def start_streaming_command(cls) -> bytes:
        """The `start_streaming` command."""
        ...

    @classmethod
    def stop_streaming_command(cls) -> bytes:
        """The `stop_streaming` command"""
        ...

    @classmethod
    def set_baudrate_command(cls, baudrate: int) -> bytes:
        ...

    @classmethod
    def parse_message(cls, header: dict[str, Any], payload: bytes) -> Message:
        """Parses any supported Message given a header and a payload"""
        ...



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtGui, QtWidgets

import pyqtgraph as pg

import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser(num_sens=1).parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = args.sensors
    config.update_rate = 30

    session_info = client.setup_session(config)

    start = session_info["range_start_m"]
    length = session_info["range_length_m"]
    num_depths = session_info["data_length"]
    step_length = session_info["step_length_m"]
    depths = np.linspace(start, start + length, num_depths)
    num_hist = 2 * int(round(config.update_rate))
    hist_data = np.zeros([num_hist, depths.size])
    smooth_max = et.utils.SmoothMax(config.update_rate)

    app = QtWidgets.QApplication([])
    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")
    pg.setConfigOptions(antialias=True)
    win = pg.GraphicsLayoutWidget()
    win.setWindowTitle("Acconeer PyQtGraph example")

    env_plot = win.addPlot(title="Envelope")
    env_plot.showGrid(x=True, y=True)
    env_plot.setLabel("bottom", "Depth (m)")
    env_plot.setLabel("left", "Amplitude")
    env_curve = env_plot.plot(pen=pg.mkPen("k", width=2))

    win.nextRow()
    hist_plot = win.addPlot()
    hist_plot.setLabel("bottom", "Time (s)")
    hist_plot.setLabel("left", "Depth (m)")
    hist_image_item = pg.ImageItem()
    tr = QtGui.QTransform()
    tr.translate(-2, start)
    tr.scale(2 / num_hist, step_length)
    hist_image_item.setTransform(tr)
    hist_plot.addItem(hist_image_item)

    # Get a nice colormap from matplotlib
    hist_image_item.setLookupTable(et.utils.pg_mpl_cmap("viridis"))

    win.show()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    win.closeEvent = lambda _: interrupt_handler.force_signal_interrupt()
    print("Press Ctrl-C to end session")

    client.start_session()

    while not interrupt_handler.got_signal:
        info, data = client.get_next()

        hist_data = np.roll(hist_data, -1, axis=0)
        hist_data[-1] = data

        env_curve.setData(depths, data)
        env_plot.setYRange(0, smooth_max.update(data))
        hist_image_item.updateImage(hist_data, levels=(0, np.max(hist_data) * 1.05))

        app.processEvents()

    print("Disconnecting...")
    app.closeAllWindows()
    client.disconnect()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from acconeer.exptool.a121._core.peripherals.communication.message import Message

from .parse_error import ParseError


class ErrorMessageHeader(te.TypedDict):
    status: te.Literal["error"]
    message: str


@attrs.frozen
class ErroneousMessage(Message):
    message: str

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> ErroneousMessage:
        t.cast(ErrorMessageHeader, header)

        if header["status"] == "error":
            return cls(header["message"])

        raise ParseError



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401
from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities.configs.sensor_config import SensorConfig
from acconeer.exptool.a121.algo.presence import Detector, DetectorConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    detector_config = DetectorConfig(
        start_m=1.0,
        end_m=3.0,
    )

    detector = Detector(client=client, sensor_id=1, detector_config=detector_config)
    detector.start()

    pg_updater = PGUpdater(
        detector_config,
        detector._get_sensor_config(detector_config),
        detector.estimated_frame_rate,
    )
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        detector_result = detector.get_next()
        s = "Presence! " if detector_result.presence_detected else "No presence. "
        s += (
            f"Intra presence score {detector_result.intra_presence_score:.3f}, "
            f"inter presence score {detector_result.inter_presence_score:.3f}, "
            f"presence at {detector_result.presence_distance:.3f} m"
        )
        print(s)
        try:
            pg_process.put_data(detector_result)
        except et.PGProccessDiedException:
            break

    detector.stop()

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(
        self,
        detector_config: DetectorConfig,
        sensor_config: SensorConfig,
        estimated_frame_rate: float,
    ):
        self.detector_config = detector_config
        self.distances = np.linspace(
            detector_config.start_m, detector_config.end_m, sensor_config.num_points
        )

        self.history_length_s = 5
        self.history_length_n = int(round(self.history_length_s * estimated_frame_rate))
        self.intra_history = np.zeros(self.history_length_n)
        self.inter_history = np.zeros(self.history_length_n)

        max_num_of_sectors = max(6, self.distances.size // 3)
        self.sector_size = max(1, -(-self.distances.size // max_num_of_sectors))
        self.num_sectors = -(-self.distances.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.distances.size) // 2

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer presence detection example")

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=0,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Distance (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_plot.setVisible(False)
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)

        # Depthwise presence plot

        self.move_plot = pg.PlotItem(title="Depthwise presence")
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Distance (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        self.move_plot.setXRange(self.distances[0], self.distances[-1])
        self.intra_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(1))
        if not self.detector_config.intra_enable:
            self.intra_curve.hide()

        self.inter_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(0))
        if not self.detector_config.inter_enable:
            self.inter_curve.hide()

        self.move_smooth_max = et.utils.SmoothMax(
            self.detector_config.frame_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_plot.addItem(self.present_text_item)
        self.move_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=1,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.detector_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            intra_dashed_pen = pg.mkPen(intra_color, width=2.5, style=QtCore.Qt.DashLine)
            intra_pen = pg.mkPen(intra_color, width=2)
        else:
            intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.detector_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=1,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.detector_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            inter_dashed_pen = pg.mkPen(inter_color, width=2.5, style=QtCore.Qt.DashLine)
            inter_pen = pg.mkPen(inter_color, width=2)
        else:
            inter_pen = et.utils.pg_pen_cycler(0)
            inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.detector_config.inter_detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=2, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_plot, row=0, col=0)
        sublayout.addItem(self.sector_plot, row=0, col=1)

        self.setup_is_done = True

    def update(self, data):
        noise = data.processor_extra_result.lp_noise
        self.noise_curve.setData(self.distances, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data.presence_distance

        self.inter_curve.setData(self.distances, data.processor_extra_result.inter)
        self.intra_curve.setData(self.distances, data.processor_extra_result.intra)
        m = self.move_smooth_max.update(
            np.max(
                np.maximum(data.processor_extra_result.inter, data.processor_extra_result.intra)
            )
        )
        m = max(
            m,
            2
            * np.maximum(
                self.detector_config.intra_detection_threshold,
                self.detector_config.inter_detection_threshold,
            ),
        )
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data.presence_detected))

        self.set_present_text_y_pos(m)

        if data.presence_detected:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        # Intra presence
        move_hist_xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.intra_history = np.roll(self.intra_history, -1)
        self.intra_history[-1] = data.intra_presence_score

        m_hist = max(
            float(np.max(self.intra_history)),
            self.detector_config.intra_detection_threshold * 1.05,
        )
        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(move_hist_xs, self.intra_history)

        # Inter presence

        self.inter_history = np.roll(self.inter_history, -1)
        self.inter_history[-1] = data.inter_presence_score

        m_hist = max(
            float(np.max(self.inter_history)),
            self.detector_config.inter_detection_threshold * 1.05,
        )
        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(move_hist_xs, self.inter_history)

        # Sector

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data.presence_detected:
            index = (
                data.processor_extra_result.presence_distance_index + self.sector_offset
            ) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y):
        x_pos = self.distances[0] + (self.distances[-1] - self.distances[0]) / 2
        self.present_text_item.setPos(x_pos, 0.95 * y)
        self.not_present_text_item.setPos(x_pos, 0.95 * y)


if __name__ == "__main__":
    main()
